from collections.abc import Iterator
from dataclasses import dataclass
from typing import Self

from ...geometry.motion import Coord, HasPosProp
from ...text_tools import Cell0D
from ..text_field import Point, TextParticleField


@dataclass
class Particle(HasPosProp):
    """
    A wrapper around an index into a text particle field's particle arrays.

    Parameters
    ----------
    field : TextParticleField
        The field which the particle belongs.
    index : int
        The index of the particle in the field.

    Attributes
    ----------
    field : TextParticleField
        The field which the particle belongs.
    index : int
        The index of the particle in the field.
    cell : Cell0D
        The particle's cell.
    pos : Point
        The particle's position.

    Methods
    -------
    iter_from_field(field)
        Yield all particles from a text particle field.
    """

    field: TextParticleField
    index: int

    @property
    def cell(self) -> Cell0D:
        """The particle's cell."""
        return self.field.particle_cells[self.index]

    @cell.setter
    def cell(self, cell: Cell0D):
        self.field.particle_cells[self.index] = cell

    @property
    def pos(self) -> Point:
        """The particle's position."""
        return Point(*self.field.particle_coords[self.index].tolist())

    @pos.setter
    def pos(self, pos: Coord):
        self.field.particle_coords[self.index] = pos

    @classmethod
    def iter_from_field(cls, field: TextParticleField) -> Iterator[Self]:
        """Yield all particles from a text particle field."""
        for i in range(len(field.particle_coords)):
            yield cls(field, i)



"""Base for creating terminal applications."""

import asyncio
import sys
from abc import ABC, abstractmethod
from contextlib import redirect_stderr
from io import StringIO
from time import perf_counter
from typing import Any, Final, cast

from . import __version__
from ._rendering import terminal_render
from ._sixel import OctTree
from .colors import NEPTUNE_THEME, Color, ColorTheme
from .gadgets._root import _Root
from .gadgets.behaviors.focusable import Focusable
from .gadgets.behaviors.themable import Themable
from .gadgets.gadget import Gadget, _GadgetList
from .gadgets.graphic_field import GraphicParticleField
from .gadgets.graphics import _BLITTER_GEOMETRY, Graphics, scale_geometry
from .geometry import Point, Size
from .logging import get_logger
from .terminal import (
    Vt100Terminal,
    app_mode,
    determine_terminal_capabilities,
    get_platform_terminal,
)
from .terminal.events import (
    ColorReportEvent,
    CursorPositionReportEvent,
    Event,
    FocusEvent,
    KeyEvent,
    MouseButton,
    MouseEvent,
    PasteEvent,
    ResizeEvent,
)

__all__ = ["App", "run_gadget_as_app"]

logger = get_logger(__name__)

_CTRL_C: Final = KeyEvent("c", ctrl=True)
"""Keybind for exiting the app."""
_TAB: Final = KeyEvent("tab")
"""Keybind for focusing next focusable."""
_SHIFT_TAB: Final = KeyEvent("tab", shift=True)
"""Keybind for focusing previous focusable."""


def _pixels_to_cells(mouse_pos, terminal_size):
    y, x = mouse_pos
    h, w = terminal_size
    return Point(y // h, x // w)


class App(ABC):
    r"""
    Base for creating terminal applications.

    Parameters
    ----------
    fg_color : Color | None, default: None
        Foreground color of the root gadget. If not given, the app will try to use the
        terminal foreground.
    bg_color : Color | None, default: None
        Background color of the root gadget. If not given, the app will try to use the
        terminal background.
    title : str | None, default: None
        The terminal's title.
    inline : bool, default: False
        Whether to render app inline or in the alternate screen.
    inline_height :int, default: 10
        Height of app if rendered inline.
    color_theme : ColorTheme, default: NEPTUNE_THEME
        Color theme for :class:`batgrl.gadgets.behaviors.themable.Themable` gadgets.
    double_click_timeout : float, default: 0.5
        Max duration of a double-click.
    render_interval : float, default: 0.0
        Duration in seconds between consecutive frame renders.

    Attributes
    ----------
    fg_color : Color | None
        Foreground color of the root gadget.
    bg_color : Color
        Background color of the root gadget.
    title : str | None
        The terminal's title.
    inline : bool
        Whether to render app inline or in the alternate screen buffer.
    inline_height :int
        Height of app if rendered inline.
    color_theme : ColorTheme
        Color theme for :class:`batgrl.gadgets.behaviors.themable.Themable` gadgets.
    double_click_timeout : float
        Max duration of a double-click.
    render_interval : float
        Duration in seconds between consecutive frame renders.
    sixel_support : bool
        Whether sixel is supported.
    sixel_geometry : Size
        Current sixel geometry.
    sixel_aspect_ratio : Size
        Current sixel aspect ratio.
    root : _Root | None
        Root of gadget tree.
    children : list[Gadget]
        Alias for :attr:`root.children`.

    Methods
    -------
    set_sixel_aspect_ratio(aspect_ratio)
        Set sixel aspect ratio.
    on_start()
        Coroutine scheduled when app is run.
    run()
        Run the app.
    exit(exit_value)
        Exit the app.
    add_gadget(gadget)
        Alias for :attr:`root.add_gadget`.
    add_gadgets(\*gadgets)
        Alias for :attr:`root.add_gadgets`.
    """

    def __init__(
        self,
        *,
        fg_color: Color | None = None,
        bg_color: Color | None = None,
        title: str | None = None,
        inline: bool = False,
        inline_height: int = 10,
        color_theme: ColorTheme = NEPTUNE_THEME,
        double_click_timeout: float = 0.5,
        render_interval: float = 0.0,
    ):
        self.root: _Root | None = None
        """Root of gadget tree (only set while app is running)."""
        self.fg_color = fg_color
        """Foreground color of the root gadget."""
        self.bg_color = bg_color
        """Background color of the root gadget."""
        self.title = title
        """The terminal's title."""
        self._inline = inline
        """Whether to render app inline or in the alternate screen buffer."""
        self.inline_height = inline_height
        """Height of app if rendered inline."""
        self.color_theme = color_theme
        """Color theme for Themable gadgets."""
        self.double_click_timeout = double_click_timeout
        """Max duration of a double-click."""
        self.render_interval = render_interval
        """Duration in seconds between consecutive frame renders."""
        self._terminal: Vt100Terminal | None = None
        """Platform-specific terminal (only set while app is running)."""
        self._app_pos: Point = Point(0, 0)
        """Position of app in terminal."""
        self._exit_value: Any = None
        """Value set by ``exit(exit_value)`` and returned by ``run()``."""
        self._octree: Final = OctTree()
        """Used by renderer to quantize graphics."""

    def __repr__(self):
        bg_color = self.bg_color if self.bg_color is None else (*self.bg_color,)
        return (
            f"{type(self).__name__}(\n"
            f"    bg_color={bg_color},\n"
            f"    title={self.title!r},\n"
            f"    inline={self.inline},\n"
            f"    inline_height={self.inline_height},\n"
            f"    double_click_timeout={self.double_click_timeout},\n"
            f"    render_interval={self.render_interval},\n"
            ")"
        )

    @property
    def is_running(self) -> bool:
        """Whether app is running."""
        return self.root is not None

    @property
    def inline(self) -> bool:
        """Whether to render app inline or in the alternate screen buffer."""
        return self._inline

    @inline.setter
    def inline(self, inline: bool):
        if inline == self._inline:
            return

        self._inline = inline
        if self._terminal is None:
            return

        root = cast(_Root, self.root)
        if inline:
            self._terminal.exit_alternate_screen()
            self._scroll_inline()
        else:
            self._terminal.erase_in_display()
            self._terminal.enter_alternate_screen()
            self._app_pos = Point(0, 0)
            root.size = self._terminal.get_size()

    def _scroll_inline(self) -> None:
        """Ensure inline mode has enought vertical space in terminal."""
        if self._terminal is None:
            return

        height = min(self.inline_height, self._terminal.get_size().height)
        self._terminal.write(b"\x0a" * height)  # Feed lines (may scroll).
        self._terminal.write(b"\x1b[%dF" % height)  # Move cursor back up.
        self._terminal.request_cursor_position_report()

    @property
    def inline_height(self) -> int:
        """
        Height of app if rendered inline.

        ``inline_height`` will be clipped to terminal height if too great.
        """
        return self._inline_height

    @inline_height.setter
    def inline_height(self, inline_height: int):
        self._inline_height = inline_height
        if self.inline and self.root is not None:
            terminal = cast(Vt100Terminal, self._terminal)
            height, _ = terminal.get_size()
            self.root.height = min(inline_height, height)

    @property
    def color_theme(self) -> ColorTheme:
        """Color theme for themable gadgets."""
        return Themable.color_theme

    @color_theme.setter
    def color_theme(self, color_theme: ColorTheme):
        Themable.color_theme = color_theme

        if self.root is not None:
            for gadget in self.root.walk():
                if isinstance(gadget, Themable):
                    gadget.update_theme()

    @property
    def fg_color(self) -> Color | None:
        """
        Foreground color of the root gadget.

        If set to ``None``, the terminal foreground color will be queried. If the
        terminal reports the foreground color, then :attr:``fg_color`` will be updated
        to the terminal's foreground color.
        """
        return self._fg_color

    @fg_color.setter
    def fg_color(self, fg_color: Color | None):
        self._fg_color = fg_color

        if self.root is None:
            return

        if fg_color is None:
            terminal = cast(Vt100Terminal, self._terminal)
            terminal.request_foreground_color()
        else:
            self.root._cell["fg_color"] = fg_color

    @property
    def bg_color(self) -> Color | None:
        """
        Background color of the root gadget.

        If set to ``None``, the terminal background color will be queried. If the
        terminal reports the background color, then :attr:``bg_color`` will be updated
        to the terminal's background color.
        """
        return self._bg_color

    @bg_color.setter
    def bg_color(self, bg_color: Color | None):
        self._bg_color = bg_color
        if self.root is None:
            return

        if bg_color is None:
            terminal = cast(Vt100Terminal, self._terminal)
            terminal.request_background_color()
        else:
            self.root.bg_color = bg_color

    @property
    def sixel_support(self) -> bool:
        """
        Whether sixel is supported.

        Will return ``False`` before app has run.
        """
        return Graphics._sixel_support

    @property
    def sixel_geometry(self) -> Size:
        """Current sixel geometry."""
        return _BLITTER_GEOMETRY["sixel"]

    @property
    def sixel_aspect_ratio(self) -> Size:
        """Current sixel aspect ratio."""
        return Graphics._sixel_aspect_ratio

    def set_sixel_aspect_ratio(self, aspect_ratio: Size) -> None:
        """
        Set sixel aspect ratio.

        Parameters
        ----------
        aspect_ratio : Size
            The desired aspect ratio. Aspect width must be 1 and aspect height must
            divide sixel geometry height.
        """
        h, w = aspect_ratio
        if w != 1 or _BLITTER_GEOMETRY["sixel"].height % h:
            raise ValueError(f"Unsupported aspect ratio: {aspect_ratio}.")
        Graphics._sixel_aspect_ratio = Size(h, w)
        if self.root is not None:
            for gadget in self.root.walk_reverse():
                if (
                    isinstance(gadget, (Graphics, GraphicParticleField))
                    and gadget.blitter == "sixel"
                ):
                    gadget.on_size()
            self.root.on_size()

    @abstractmethod
    async def on_start(self):
        """Coroutine scheduled when app is run."""

    def run(self) -> Any:
        """Run the app."""
        logger.info("batgrl v%s", __version__)
        logger.info("Starting app")
        tmp_stderr = StringIO()

        try:
            with redirect_stderr(tmp_stderr):
                asyncio.run(self._run_async())
        except asyncio.CancelledError:
            pass
        finally:
            sys.stderr.write(tmp_stderr.getvalue())

        logger.info("Exiting app")
        exit_value = self._exit_value
        self._exit_value = None
        return exit_value

    def exit(self, exit_value: Any = None) -> None:
        """
        Exit the app.

        Parameters
        ----------
        exit_value : Any, default: None
            Value returned by ``run()``.
        """
        self._exit_value = exit_value
        if self.root is not None:
            self.root.destroy()
            self.root = None
        self._terminal = None

        try:
            tasks = asyncio.all_tasks()
        except RuntimeError:
            pass
        else:
            for task in tasks:
                task.cancel()

    async def _run_async(self):
        """Build environment, create root, and schedule app-specific tasks."""
        terminal = self._terminal = get_platform_terminal()
        last_size: Size = terminal.get_size()
        self.root = root = _Root(app=self, size=last_size)

        last_mouse_button: MouseButton = "no_button"
        last_mouse_time = perf_counter()
        last_mouse_nclicks = 0

        def determine_nclicks(mouse_event: MouseEvent) -> None:
            """Determine number of consecutive clicks for a `MouseEvent`."""
            nonlocal last_mouse_button, last_mouse_time, last_mouse_nclicks
            current_time = perf_counter()

            if mouse_event.event_type != "mouse_down":
                return

            if (
                last_mouse_button != mouse_event.button
                or current_time - last_mouse_time > self.double_click_timeout
            ):
                mouse_event.nclicks = 1
            else:
                mouse_event.nclicks = last_mouse_nclicks % 3 + 1

            last_mouse_button = mouse_event.button
            last_mouse_nclicks = mouse_event.nclicks
            last_mouse_time = current_time

        def event_handler(events: list[Event]) -> None:
            """Handle input events."""
            nonlocal last_size

            for event in events:
                if isinstance(event, KeyEvent):
                    if event == _CTRL_C:
                        self.exit()
                        return
                    if not root.dispatch_key(event):
                        if event == _TAB:
                            Focusable.focus_next()
                        elif event == _SHIFT_TAB:
                            Focusable.focus_previous()
                elif isinstance(event, MouseEvent):
                    determine_nclicks(event)
                    if event.pos is None:
                        event.pos = _pixels_to_cells(event.pixels_pos, last_size)
                    event.pos -= self._app_pos
                    if event.pixels_pos is not None:
                        event.pixels_pos -= scale_geometry("sixel", self._app_pos)
                    root.dispatch_mouse(event)
                elif isinstance(event, PasteEvent):
                    root.dispatch_paste(event)
                elif isinstance(event, FocusEvent):
                    root.dispatch_terminal_focus(event)
                elif isinstance(event, ResizeEvent):
                    if event.size == last_size:
                        # Sometimes spurious resize events can appear such as when a
                        # terminal first enables VT100 processing or when
                        # entering/exiting the alternate screen buffer.
                        continue
                    last_size = event.size
                    if self.inline:
                        terminal.erase_in_display()
                        self._scroll_inline()
                    else:
                        root.size = event.size
                elif isinstance(event, CursorPositionReportEvent):
                    if self.inline:
                        height, width = last_size
                        root.size = Size(
                            min(self.inline_height, height), width - event.pos.x
                        )

                        # Needs to be manually set in case root.size hasn't changed.
                        root._resized = True
                        self._app_pos = event.pos
                elif isinstance(event, ColorReportEvent):
                    if event.kind == "fg":
                        self.fg_color = event.color
                    else:
                        self.bg_color = event.color

        async def auto_render():
            """Render screen every ``render_interval`` seconds."""
            while True:
                await asyncio.sleep(self.render_interval)

                if terminal.expect_dsr():
                    continue

                resized = root._resized
                root._render()
                terminal_render(
                    resized,
                    self,
                    root,
                    Graphics._sixel_aspect_ratio,
                )

        with app_mode(terminal, event_handler):
            (
                Graphics._sixel_support,
                _BLITTER_GEOMETRY["sixel"],
            ) = await determine_terminal_capabilities(terminal)
            root.on_size()  # Make cell and graphics arrays.

            if self.title is not None:
                terminal.set_title(self.title)

            if self.fg_color is None:
                terminal.request_foreground_color()
            else:
                self.root._cell["fg_color"] = self.fg_color

            if self.bg_color is None:
                terminal.request_background_color()
            else:
                self.root.bg_color = self.bg_color

            if self.inline:
                self._scroll_inline()
            else:
                terminal.enter_alternate_screen()

            await asyncio.gather(self.on_start(), auto_render())

    def add_gadget(self, gadget: Gadget) -> None:
        """
        Alias for :attr:`root.add_gadget`.

        Parameters
        ----------
        gadget : Gadget
            A gadget to add as a child to the root gadget.
        """
        if self.root is not None:
            self.root.add_gadget(gadget)

    def add_gadgets(self, *gadgets: Gadget) -> None:
        r"""
        Alias for :attr:`root.add_gadgets`.

        Parameters
        ----------
        \*gadgets : Gadget
            Gadgets to add as children to the root gadget.
        """
        if self.root is not None:
            self.root.add_gadgets(*gadgets)

    @property
    def children(self) -> _GadgetList[Gadget] | None:
        """Alias for :attr:`root.children`."""
        if self.root is not None:
            return self.root.children


def run_gadget_as_app(
    gadget: Gadget,
    *,
    fg_color: Color | None = None,
    bg_color: Color | None = None,
    title: str | None = None,
    inline: bool = False,
    inline_height: int = 10,
    color_theme: ColorTheme = NEPTUNE_THEME,
    double_click_timeout: float = 0.5,
    render_interval: float = 0.0,
) -> Any:
    """
    Run a gadget as an app.

    This convenience function provided for cases when the app would only have a single
    gadget.

    Parameters
    ----------
    gadget : Gadget
        A gadget to run as an app.
    fg_color : Color | None, default: None
        Foreground color of the root gadget.
    bg_color : Color | None, default: None
        Background color of the root gadget.
    title : str | None, default: None
        The terminal's title.
    inline : bool, default: False
        Whether to render app inline or in the alternate screen.
    inline_height :int, default: 10
        Height of app if rendered inline.
    color_theme : ColorTheme, default: NEPTUNE_THEME
        Color theme for :class:`batgrl.gadgets.behaviors.themable.Themable` gadgets.
    double_click_timeout : float, default: 0.5
        Max duration of a double-click.
    render_interval : float, default: 0.0
        Duration in seconds between consecutive frame renders.
    """

    class _DefaultApp(App):
        async def on_start(self):
            self.add_gadget(gadget)

    return _DefaultApp(
        fg_color=fg_color,
        bg_color=bg_color,
        title=title,
        inline=inline,
        inline_height=inline_height,
        color_theme=color_theme,
        double_click_timeout=double_click_timeout,
        render_interval=render_interval,
    ).run()



"""Focus behavior for a gadget."""

from collections import deque
from weakref import ReferenceType, WeakSet, ref

from . import Behavior

__all__ = ["Focusable"]


class AnyFocusedProperty:
    def __get__(self, instance, owner) -> bool:
        return bool(Focusable._focused)


class Focusable(Behavior):
    """
    Focus behavior for a gadget.

    Focusable gadgets can be given "focus" by pressing tab or shift + tab
    or by clicking on them. When a gadget is focused, all focusable ancestors
    also gain focus.

    Attributes
    ----------
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.

    Methods
    -------
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    """

    __focusables: deque[ReferenceType] = deque()
    """Focusables that are part of the gadget tree."""
    _focused: WeakSet = WeakSet()
    """Focused focusables."""
    any_focused: AnyFocusedProperty = AnyFocusedProperty()
    """Whether any gadget has focus."""

    @classmethod
    def _focus(cls, step: int):
        focusables = cls.__focusables

        for _ in range(len(focusables)):
            gadget = focusables[0]()
            if gadget is None:
                focusables.popleft()
            elif (
                not gadget.is_visible or not gadget.is_enabled or gadget in cls._focused
            ):
                focusables.rotate(step)
            else:
                gadget.focus()
                break

    @classmethod
    def focus_next(cls):
        """Focus next focusable gadget."""
        cls._focus(-1)

    @classmethod
    def focus_previous(cls):
        """Focus previous focusable gadget."""
        cls._focus(1)

    @property
    def is_focused(self) -> bool:
        """Whether gadget has focus."""
        return self in Focusable._focused

    def focus(self):
        """Focus gadget."""
        if self.parent is None or not self.is_enabled or not self.is_visible:
            return

        ancestors = WeakSet(
            ancestor for ancestor in self.ancestors() if isinstance(ancestor, Focusable)
        )
        ancestors.add(self)

        focused = Focusable._focused
        Focusable._focused = ancestors

        for blurred in focused - ancestors:
            blurred.on_blur()

        for needs_focus in ancestors - focused:
            needs_focus.on_focus()

        # Try to move self to top of focusables queue.
        focus_gadgets = Focusable.__focusables
        i = 0
        while i < len(focus_gadgets):
            gadget = focus_gadgets[0]()
            if gadget is None:
                focus_gadgets.popleft()
            elif gadget is self:
                break
            else:
                focus_gadgets.rotate(-1)
                i += 1

    def blur(self):
        """Un-focus gadget."""
        if self.is_focused:
            for ancestor in self.ancestors():
                if isinstance(ancestor, Focusable):
                    ancestor.focus()
                    return

            Focusable._focused.discard(self)
            self.on_blur()

    def on_focus(self):
        """Update gadget when it gains focus."""

    def on_blur(self):
        """Update gadget when it loses focus."""

    def on_add(self):
        """Add to focusable gadgets and focus on add."""
        super().on_add()
        Focusable.__focusables.append(ref(self))
        self.focus()

    def on_remove(self):
        """Remove from focusable gadgets and blur on remove."""
        self.blur()
        Focusable.__focusables.remove(ref(self))
        super().on_remove()

    def dispatch_key(self, key_event) -> bool | None:
        """Dispatch key press only if focused."""
        return self.is_focused and super().dispatch_key(key_event)

    def dispatch_mouse(self, mouse_event) -> bool | None:
        """Focus if mouse event is handled."""
        handled = super().dispatch_mouse(mouse_event)
        if handled and not self.is_focused and self.is_visible:
            self.focus()
        return handled

    def on_mouse(self, mouse_event) -> bool | None:
        """Focus on mouse down collision and blur otherwise."""
        if mouse_event.event_type == "mouse_down" and self.is_visible:
            collides = self.collides_point(mouse_event.pos)
            if not self.is_focused and collides:
                self.focus()
            elif self.is_focused and not collides:
                self.blur()

        return super().on_mouse(mouse_event)



"""A video gadget."""

import asyncio
import atexit
import time
from pathlib import Path
from platform import uname
from typing import cast

import cv2

from ..colors import ABLACK, AColor
from ..logging import get_logger
from ..texture_tools import resize_texture
from .graphics import (
    Blitter,
    Graphics,
    Interpolation,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    scale_geometry,
)

__all__ = ["Interpolation", "Point", "Size", "Video"]

_IS_WSL: bool = uname().system == "Linux" and uname().release.endswith("Microsoft")

logger = get_logger(__name__)


class Video(Graphics):
    r"""
    A video gadget.

    Parameters
    ----------
    source : pathlib.Path | str | int
        A path to video, URL to video stream, or video capturing device (by index).
        Trying to open a video capturing device on WSL will issue a warning.
    loop : bool, default: True
        Whether to restart video after last frame.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    source : Path | str | int
        A path, URL, or capturing device (by index) of the video.
    loop : bool
        Whether to restart video after last frame.
    is_device : bool
        Whether video is from a video capturing device.
    texture : RGBA_2D
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    play()
        Play the video. Returns a task.
    pause()
        Pause the video.
    seek()
        Seek to certain time (in seconds) in the video.
    stop()
        Stop the video.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        source: Path | str | int,
        loop: bool = True,
        default_color: AColor = ABLACK,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._current_frame = None
        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._resource: cv2.VideoCapture | None = None
        self._video_task = None
        self.source = source
        self.loop = loop

    @property
    def source(self) -> Path | str | int:
        """A path, URL, or capturing device (by index) of the video."""
        return self._source

    @source.setter
    def source(self, source: Path | str | int):
        self.pause()
        self._release_resource()
        if isinstance(source, Path) and not source.exists():
            raise FileNotFoundError(str(source))
        self._source = source
        self._load_resource()

    @property
    def is_device(self):
        """Return true if source is a video capturing device."""
        return isinstance(self._source, int)

    def _load_resource(self):
        source = self.source

        if _IS_WSL and self.is_device:
            # Because WSL doesn't support most USB devices (yet?), and trying to open
            # one with cv2 will pollute the terminal with cv2 errors, we don't attempt
            # to open a device in this case.
            logger.info("Device not available on WSL.")
            self._resource = None
            return

        if isinstance(source, Path):
            source = str(source.absolute())

        self._resource = cv2.VideoCapture(source)
        atexit.register(self._resource.release)

    def _release_resource(self):
        if self._resource is not None:
            self._resource.release()
            atexit.unregister(self._resource.release)
            self._resource = None
            self._current_frame = None
            self.clear()

    def _time_delta(self) -> float:
        resource = cast(cv2.VideoCapture, self._resource)
        return time.perf_counter() - resource.get(cv2.CAP_PROP_POS_MSEC) / 1000

    def _display_current_frame(self):
        h, w = self.size
        if self._current_frame is None or h == 0 or w == 0:
            return

        resize_texture(
            self._current_frame,  # type: ignore
            scale_geometry(self._blitter, self._size),
            self._interpolation,
            out=self.texture,
        )

    async def _play_video(self):
        if self._resource is None:
            return

        self._start_time = self._time_delta()

        while True:
            if not self._resource.grab():
                if self.loop:
                    self.seek(0)
                else:
                    self._current_frame = None
                    self.clear()
                    return

            if self.is_device:
                seconds_ahead = 0
            elif (seconds_ahead := self._start_time - self._time_delta()) < 0:
                continue

            try:
                await asyncio.sleep(seconds_ahead)
            except asyncio.CancelledError:
                return

            _, frame = self._resource.retrieve()
            self._current_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)
            self._display_current_frame()

    def on_size(self):
        """Resize current frame on resize."""
        super().on_size()
        self._display_current_frame()

    def on_remove(self):
        """Pause video and release resource on remove."""
        super().on_remove()
        self.pause()
        self._release_resource()

    def pause(self):
        """Pause video."""
        if self._video_task is not None:
            self._video_task.cancel()

    def play(self) -> asyncio.Task:
        """
        Play video.

        Returns
        -------
        asyncio.Task
            The task that plays the video.
        """
        self.pause()

        if self._resource is None:
            self._load_resource()

        self._video_task = asyncio.create_task(self._play_video())
        return self._video_task

    def seek(self, time: float):
        """If supported, seek to certain time (in seconds) in the video."""
        if self._resource is not None and not self.is_device:
            self._resource.set(cv2.CAP_PROP_POS_MSEC, time * 1000)
            self._resource.grab()
            self._start_time = self._time_delta()

    def stop(self):
        """Stop video."""
        self.pause()
        self.seek(0)
        self._current_frame = None
        self.clear()



from libc.stdint cimport uint32_t


cdef packed struct Cell:
    uint32_t ord
    unsigned char style
    unsigned char[3] fg_color
    unsigned char[3] bg_color


cdef packed struct BraillePixel:
    uint32_t ord
    double[4] total_fg
    uint32_t ncolors



"""A text gadget."""

from typing import Any, Final, Literal, NamedTuple

import numpy as np
from tree_sitter import QueryCursor, Tree

from .._rendering import text_render
from ..array_types import RGBM_2D, Cell0D, Cell2D, Enum2D, Unicode2D
from ..colors import Color, Neptune, SyntaxHighlightTheme
from ..queries import Highlighter, TSPoint, get_highlighter
from ..text_tools import (
    Style,
    _parse_batgrl_md,
    _text_to_cells,
    _write_cells_to_canvas,
    add_text,
    egc_chr,
    egc_ord,
    new_cell,
)
from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    bindable,
    clamp,
)

__all__ = [
    "Border",
    "Point",
    "Size",
    "Text",
    "add_text",
]

Border = Literal[
    "light",
    "heavy",
    "double",
    "curved",
    "ascii",
    "outer",
    "inner",
    "thick",
    "dashed",
    "dashed_2",
    "dashed_3",
    "heavy_dashed",
    "heavy_dashed_2",
    "heavy_dashed_3",
    "near",
    "octant",
    "octant_rounded",
    "mcgugan_tall",
    "mcgugan_wide",
    "sextant",
    "sextant_rounded",
]
"""Border styles for :meth:`batgrl.text_gadget.Text.add_border`."""

_BORDERS: Final[dict[Border, str]] = {
    "light": "┌┐││──└┘",
    "heavy": "┏┓┃┃━━┗┛",
    "double": "╔╗║║══╚╝",
    "curved": "╭╮││──╰╯",
    "ascii": "++||--++",
    "outer": "▛▜▌▐▀▄▙▟",
    "inner": "▗▖▐▌▄▀▝▘",
    "thick": "████▀▄██",
    "dashed": "┌┐╎╎╌╌└┘",
    "dashed_2": "┌┐┆┆┄┄└┘",
    "dashed_3": "┌┐┊┊┈┈└┘",
    "heavy_dashed": "┏┓╏╏╍╍┗┛",
    "heavy_dashed_2": "┏┓┇┇┅┅┗┛",
    "heavy_dashed_3": "┏┓┋┋┉┉┗┛",
    "near": "  ▕▏▁▔  ",
    "octant": "𜵊𜶘▌▐🮂▂𜷀𜷕",
    "octant_rounded": "𜵉𜶗▌▐🮂▂𜶅𜵛",
    "mcgugan_tall": "▕▏▕▏▔▁▕▏",
    "mcgugan_wide": "▁▁▏▕▁▔▔▔",
    "sextant": "🬕🬨▌▐🬂🬭🬲🬷",
    "sextant_rounded": "🬔🬧▌▐🬂🬭🬣🬘",
}
"""Border characters for :meth:`batgrl.text_gadget.Text.add_border`."""


def _fill_array(sy: int, sx: int, ey: int, ex: int, arr: np.ndarray, value: Any):
    """Fill ``arr`` from ``(sy, sx)`` to ``(ey, ex)`` with ``value``."""
    if sy == ey:
        arr[sy : sy + 1, sx:ex] = value
    else:
        arr[sy : sy + 1, sx:] = value
        arr[sy + 1 : ey] = value
        arr[ey : ey + 1, :ex] = value


class _InjectionRange(NamedTuple):
    start_row: int
    start_byte_offset: int
    start_column: int
    end_row: int
    end_byte_offset: int
    end_column: int


class Text(Gadget):
    r"""
    A gadget to display arbitrary styled text.

    Parameters
    ----------
    default_cell : Cell0D | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 0.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    canvas : Cell2D
        The array of characters for the gadget.
    chars : Unicode2D
        Return a view of the ords field of the canvas as 1-character unicode strings.
    default_cell : Cell0D
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(language, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        default_cell: Cell0D | str = " ",
        alpha: float = 0.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        size = self.size
        self.default_cell = default_cell
        self.canvas: Cell2D = np.full(size, self.default_cell)
        self.alpha = alpha
        self._injection_range: _InjectionRange | None = None
        """
        The range of some injected languages (such as, e.g., a regex string in python
        code) in the canvas when syntax highlighting.

        Set by `_highlight`, ``_injection_range`` modifies the behavior of both
        ``_tree_sitter_point_to_pos`` and ``_tree_sitter_read_canvas``.
        """

    @property
    def chars(self) -> Unicode2D:
        """
        Return a view of the ords field of the canvas as 1-character unicode strings.

        Warning
        -------
        This view may raise a ``RuntimeError`` if extended grapheme clusters are stored
        in the canvas as the egc flag is greater than sys.maxunicode.
        """
        return self.canvas["ord"].view("<U1")

    @property
    def default_cell(self) -> Cell0D:
        """Default character for text canvas."""
        return self._default_cell

    @default_cell.setter
    def default_cell(self, cell: Cell0D | str):
        if isinstance(cell, str):
            cell = new_cell(ord=egc_ord(cell))
        self._default_cell = cell

    @property
    def default_fg_color(self) -> Color:
        """Foreground color of default character."""
        return Color(*self._default_cell["fg_color"].tolist())

    @default_fg_color.setter
    def default_fg_color(self, default_fg_color: Color):
        self._default_cell["fg_color"] = default_fg_color

    @property
    def default_bg_color(self) -> Color:
        """Background color of default character."""
        return Color(*self._default_cell["bg_color"].tolist())

    @default_bg_color.setter
    def default_bg_color(self, default_bg_color: Color):
        self._default_cell["bg_color"] = default_bg_color

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)

    def on_size(self):
        """Resize canvas preserving as much content as possible."""
        old_canvas = self.canvas
        old_h, old_w = old_canvas.shape

        h, w = self._size

        copy_h = min(old_h, h)
        copy_w = min(old_w, w)

        self.canvas = np.full((h, w), self.default_cell)
        self.canvas[:copy_h, :copy_w] = old_canvas[:copy_h, :copy_w]

    def add_border(
        self,
        style: Border = "light",
        bold: bool = False,
        fg_color: Color | None = None,
        bg_color: Color | None = None,
    ):
        """
        Add a text border (gadget height and width must be at least 2).

        Parameters
        ----------
        style : Border, default: "light"
            Style of border. Default style uses light box-drawing characters.
        bold : bool, default: False
            Whether the border is bold.
        fg_color : Color | None, default: None
            Foreground color of border if not None.
        bg_color : Color | None, default: None
            Background color of border if not None.
        """
        if self.height < 2 or self.width < 2:
            return

        slices = [
            np.s_[0, 0],
            np.s_[0, -1],
            np.s_[1:-1, 0],
            np.s_[1:-1, -1],
            np.s_[0, 1:-1],
            np.s_[-1, 1:-1],
            np.s_[-1, 0],
            np.s_[-1, -1],
        ]
        for s, border in zip(slices, _BORDERS[style]):
            self.chars[s] = border
            if bold:
                self.canvas["style"][s] = Style.BOLD

        if fg_color is not None:
            self.canvas["fg_color"][[0, -1]] = fg_color
            self.canvas["fg_color"][:, [0, -1]] = fg_color
        if bg_color is not None:
            self.canvas["bg_color"][[0, -1]] = bg_color
            self.canvas["bg_color"][:, [0, -1]] = bg_color

    def _tree_sitter_point_to_pos(
        self, point: tuple[int, int], prev: tuple[int, int, int] | None = None
    ) -> Point:
        """
        Convert a tree sitter (row, byte_offset) to a point in the canvas.

        ``prev`` can reuse a previous `point_to_pos` calculation and has the form
        (prev_row, prev_byte_offset, prev_x).
        """
        y, byte_offset = point
        x = nbytes = 0
        end_column = self.width
        if self._injection_range is not None:
            y += self._injection_range.start_row
            if y == self._injection_range.start_row:
                x = self._injection_range.start_column
            if y == self._injection_range.end_row:
                end_column = self._injection_range.end_column
            elif y > self._injection_range.end_row:
                return Point(self._injection_range.end_row + 1, 0)
        elif y >= self.height:
            return Point(self.height, 0)

        if prev is not None and y == prev[0] and prev[1] < byte_offset:
            _, nbytes, x = prev

        while nbytes < byte_offset:
            nbytes += len(egc_chr(self.canvas["ord"][y, x]).encode())
            x += 1

            if x >= end_column:
                return Point(y, end_column)

        return Point(y, x)

    def _tree_sitter_read_canvas(self, _, point: tuple[int, int]) -> bytes:
        y, x = self._tree_sitter_point_to_pos(point)
        if y >= self.height:
            return b""

        end_column = None
        if self._injection_range is not None:
            if y > self._injection_range.end_row:
                return b""

            if y == self._injection_range.end_row:
                end_column = self._injection_range.end_column

        ords = self.canvas["ord"][y, x:end_column].tolist()
        return "".join(egc_chr(ord_) for ord_ in ords).rstrip().encode() + b"\n"

    def _highlight(
        self,
        theme: SyntaxHighlightTheme,
        highlighter: Highlighter,
        syntax_tree: Tree,
        point_range: tuple[TSPoint, TSPoint] | None = None,
    ) -> None:
        highlights_cursor = QueryCursor(highlighter.highlights)
        if point_range is not None:
            highlights_cursor.set_point_range(*point_range)
        elif self._injection_range is None:
            self.canvas["fg_color"] = theme.default_fg
            self.canvas["bg_color"] = theme.default_bg

        highlights_matches = highlights_cursor.matches(syntax_tree.root_node)
        # Sorting matches by query priority. Later queries have greater priority.
        highlights_matches.sort(key=lambda tup: tup[0])
        prev = (-1, 0, 0)
        tokens = (
            (highlight, *node.start_point, *node.end_point)
            for _, captures in highlights_matches
            for highlight, nodes in captures.items()
            for node in nodes
        )
        for highlight, sy, sbo, ey, ebo in tokens:
            if highlight not in theme.highlights:
                continue

            sy, sx = self._tree_sitter_point_to_pos((sy, sbo), prev=prev)
            ey, ex = self._tree_sitter_point_to_pos((ey, ebo), prev=(sy, sbo, sx))
            prev = ey, ebo, ex

            style, fg_color = theme.highlights[highlight]
            _fill_array(sy, sx, ey, ex, self.canvas["style"], style)
            # Un-style whitespace:
            self.canvas["style"][self.canvas["ord"] == 0x20] = 0
            if fg_color is not None:
                _fill_array(sy, sx, ey, ex, self.canvas["fg_color"], fg_color)

        if highlighter.injections is None:
            return

        injections_cursor = QueryCursor(highlighter.injections)
        if point_range is not None:
            injections_cursor.set_point_range(*point_range)

        prev_injection_range = self._injection_range

        injections_matches = injections_cursor.matches(syntax_tree.root_node)
        injections = (
            (language, *node.start_point, *node.end_point)
            for _, captures in injections_matches
            for language, nodes in captures.items()
            for node in nodes
        )
        for language, sy, sbo, ey, ebo in injections:
            injection_highlighter = get_highlighter(language)
            if injection_highlighter is None:
                continue

            sy, sx = self._tree_sitter_point_to_pos((sy, sbo))
            ey, ex = self._tree_sitter_point_to_pos((ey, ebo), prev=(sy, sbo, sx))
            self._injection_range = _InjectionRange(sy, sbo, sx, ey, ebo, ex)
            injected_syntax_tree = injection_highlighter.parser.parse(
                self._tree_sitter_read_canvas
            )
            self._highlight(theme, injection_highlighter, injected_syntax_tree)
            self._injection_range = prev_injection_range

    def add_syntax_highlighting(
        self, language: str, theme: SyntaxHighlightTheme = Neptune
    ):
        """
        Add syntax highlighting to current text in canvas.

        Parameters
        ----------
        language : str
            The language for syntax highlighting.
        theme : SyntaxHighlightTheme, default: Neptune
            A theme to use for syntax highlighting.
        """
        highlighter = get_highlighter(language)
        if highlighter is None:
            return

        syntax_tree = highlighter.parser.parse(self._tree_sitter_read_canvas)
        self._highlight(theme, highlighter, syntax_tree)

    def add_str(
        self,
        str: str,
        *,
        pos: Pointlike = Point(0, 0),
        fg_color: Color | None = None,
        bg_color: Color | None = None,
        markdown: bool = False,
        truncate_str: bool = False,
    ):
        """
        Add a single line of text to the canvas at position `pos`.

        If `markdown` is true, text can be styled using batgrl markdown. The syntax is:
        - italic: `*this is italic text*`
        - bold: `**this is bold text**`
        - strikethrough: `~~this is strikethrough text~~`
        - underlined: `__this is underlined text__`
        - overlined: `^^this is overlined text^^`

        Parameters
        ----------
        str : str
            A single line of text to add to canvas.
        pos : Pointlike, default: Point(0, 0)
            Position of first character of string. Negative coordinates position
            from the right or bottom of canvas (like negative indices).
        fg_color : Color | None, default: None
            Foreground color of text.
        bg_color : Color | None, default: None
            Background color of text.
        markdown : bool, default: False
            Whether to parse text for batgrl markdown.
        truncate_str : bool, default: False
            If false, an `IndexError` is raised if the text would not fit on canvas.

        See Also
        --------
        text_tools.add_text : Add multiple lines of text to a view of a canvas.
        """
        y, x = pos
        add_text(
            self.canvas[y, x:],
            str,
            fg_color=fg_color,
            bg_color=bg_color,
            markdown=markdown,
            truncate_text=truncate_str,
        )

    def set_text(
        self,
        text: str,
        *,
        fg_color: Color | None = None,
        bg_color: Color | None = None,
        markdown: bool = False,
    ):
        """
        Resize gadget to fit text, erase canvas, then fill canvas with text.

        If `markdown` is true, text can be styled using batgrl markdown. The syntax is:
        - italic: `*this is italic text*`
        - bold: `**this is bold text**`
        - strikethrough: `~~this is strikethrough text~~`
        - underlined: `__this is underlined text__`
        - overlined: `^^this is overlined text^^`

        Parameters
        ----------
        text : str
            Text to add to canvas.
        fg_color : Color | None, default: None
            Foreground color of text.
        bg_color : Color | None, default: None
            Background color of text.
        markdown : bool, default: False
            Whether to parse text for batgrl markdown.

        See Also
        --------
        text_tools.add_text : Add multiple lines of text to a view of a canvas.
        """
        self.size, lines = _parse_batgrl_md(text) if markdown else _text_to_cells(text)
        self.clear()
        _write_cells_to_canvas(lines, self.canvas, fg_color, bg_color)

    def clear(self):
        """Fill canvas with default cell."""
        self.canvas[:] = self.default_cell

    def shift(self, n: int = 1):
        """
        Shift content in canvas up (or down in case of negative `n`).

        Rows at the bottom (or top) will be filled with the default cell.
        """
        if n > 0:
            self.canvas[:-n] = self.canvas[n:]
            self.canvas[-n:] = self.default_cell
        elif n < 0:
            self.canvas[-n:] = self.canvas[:n]
            self.canvas[:-n] = self.default_cell

    def _render(self, cells: Cell2D, graphics: RGBM_2D, kind: Enum2D) -> None:
        """Render visible region of gadget."""
        text_render(
            cells,
            graphics,
            kind,
            self.absolute_pos,
            self._is_transparent,
            self.canvas,
            self.alpha,
            self._region,
        )



"""Basic geometry functions and types."""

from __future__ import annotations

from typing import NamedTuple, Self

import numpy as np

from ..array_types import Coords

__all__ = [
    "Point",
    "Pointlike",
    "Size",
    "Sizelike",
    "clamp",
    "lerp",
    "normalize",
    "points_on_circle",
    "rect_slice",
    "round_down",
]


def clamp[T: (float, int)](value: T, min: T | None, max: T | None) -> T:
    """
    If ``value`` is less than ``min``, returns ``min``; else if ``max`` is less than
    ``value``, returns ``max``; else returns ``value``. A one-sided clamp is possible
    by setting ``min`` or ``max`` to ``None``.

    Parameters
    ----------
    value : (float, int)
        Value to clamp.
    min : (float, int) | None
        Minimum of clamped value.
    max : (float, int) | None
        Maximum of clamped value.

    Returns
    -------
    (float, int)
        A value between `min` and `max`, inclusive.
    """
    if min is not None and value < min:
        return min

    if max is not None and value > max:
        return max

    return value


def lerp(a: float, b: float, p: float) -> float:
    """Linear interpolation of `a` to `b` with proportion `p`."""
    return (1.0 - p) * a + p * b


def normalize(p: float, start: float, end: float) -> float:
    """Return ``p`` as a proportion of the interval from ``start`` to ``end``."""
    return (p - start) / (end - start)


def points_on_circle(
    n: int,
    radius: float = 1.0,
    center: tuple[float, float] = (0.0, 0.0),
    offset: float = 0.0,
) -> Coords:
    """
    Return `n` points on a circle.

    Parameters
    ----------
    n : int
        Number of points on a circle.
    radius : float, default: 1.0
        Radius of circle.
    center : tuple[float, float], default: (0.0, 0.0)
        Center of circle.
    offset : float, default: 0.0
        Rotate output points by `offset` radians.

    Returns
    -------
    Coords
        An `(n, 2)`-shaped array of evenly-spaced points on a circle.
    """
    angles = np.linspace(0, 2 * np.pi, endpoint=False, num=n) + offset
    return radius * np.stack([np.sin(angles), np.cos(angles)]).T + center


def rect_slice(
    pos: Pointlike, size: Sizelike
) -> tuple[slice[int, int, None], slice[int, int, None]]:
    """
    Return slices for indexing a rect in a numpy array.

    Parameters
    ----------
    pos : Pointlike
        Position of rect.
    size : Sizelike
        Size of rect.

    Returns
    -------
    tuple[slice[int, int, None], slice[int, int, None]]
        Slices that index a rect in a numpy array.
    """
    y, x = pos
    h, w = size
    return slice(y, y + h), slice(x, x + w)


def round_down(n: float) -> int:
    """
    Like the built-in `round`, but always rounds down.

    Used instead of `round` for smoother geometry.
    """
    i, r = divmod(n, 1)
    if r <= 0.5:
        return int(i)
    return int(i) + 1


class Point(NamedTuple):
    """
    A 2-d point.

    Note that y-coordinate is before x-coordinate. This convention is used so that the
    2-d arrays that underly a gadget's data can be directly indexed with the point.

    Parameters
    ----------
    y : int
        y-coordinate of point.
    x : int
        x-coordinate of point.

    Attributes
    ----------
    y : int
        y-coordinate of point.
    x : int
        x-coordinate of point.
    """

    y: int
    """y-coordinate of point."""
    x: int
    """x-coordinate of point."""

    def __add__(self, other: Pointlike) -> Self:  # type: ignore
        y, x = self
        oy, ox = other

        return type(self)(y + oy, x + ox)

    def __radd__(self, other: Pointlike) -> Self:
        oy, ox = other
        y, x = self
        return type(self)(oy + y, ox + x)

    def __sub__(self, other: Pointlike) -> Self:
        y, x = self
        oy, ox = other
        return type(self)(y - oy, x - ox)

    def __rsub__(self, other: Pointlike) -> Self:
        oy, ox = other
        y, x = self
        return type(self)(oy - y, ox - x)

    def __neg__(self) -> Self:
        y, x = self
        return type(self)(-y, -x)


class Size(NamedTuple):
    """
    A rectangular area.

    Parameters
    ----------
    height : int
        Height of area.
    width : int
        Width of area.

    Attributes
    ----------
    height : int
        Height of area.
    width : int
        Width of area.
    rows : int
        Alias for height.
    columns : int
        Alias for width.
    center : Point
        Center of area.
    """

    height: int
    """Height of area."""
    width: int
    """Width of area."""

    @property
    def rows(self) -> int:
        """Alias for height."""
        return self.height

    @property
    def columns(self) -> int:
        """Alias for width."""
        return self.width

    def __contains__(self, point: Pointlike) -> bool:  # type: ignore
        """Whether a point is within the rectangle."""
        y, x = point
        h, w = self
        return 0 <= y < h and 0 <= x < w

    @property
    def center(self) -> Point:
        """Center of area."""
        h, w = self
        return Point(h // 2, w // 2)


type Pointlike = Point | tuple[int, int]
type Sizelike = Size | tuple[int, int]



"""A file chooser gadget."""

import sys
from collections.abc import Callable
from pathlib import Path
from typing import cast

from uwcwidth import wcswidth

from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
)
from .scroll_view import ScrollView
from .tree_view import TreeView, TreeViewNode

__all__ = ["FileChooser", "Point", "Size"]

_FILE_PREFIX = "  📄 "
_FOLDER_PREFIX = "▶ 📁 "
_NESTED_PREFIX = "  "
_OPEN_FOLDER_PREFIX = "▼ 📂 "

if sys.platform == "win32":
    from ctypes import windll

    def _is_hidden(path: Path) -> bool:
        attrs = windll.kernel32.GetFileAttributesW(str(path.absolute()))
        # https://docs.microsoft.com/en-us/windows/win32/fileio/file-attribute-constants
        is_hidden = 0x2 | 0x4
        return attrs != -1 and bool(attrs & is_hidden)

else:

    def _is_hidden(path: Path) -> bool:
        return path.stem.startswith(".")


class _ParentDirectory(TreeViewNode):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.label = ".."
        self.set_text(self.label)

    def on_mouse(self, mouse_event):
        if (
            mouse_event.nclicks == 2
            and cast(TreeView, self.parent).selected_node is self
            and self.collides_point(mouse_event.pos)
        ):
            for ancestor in self.ancestors():
                if isinstance(ancestor, FileChooser):
                    break
            else:
                return False

            root_dir = ancestor.root_dir
            root_dir_parent = root_dir.parent
            if root_dir != root_dir_parent:
                ancestor.root_dir = root_dir_parent
            return True

        return super().on_mouse(mouse_event)


class _FileViewNode(TreeViewNode):
    def __init__(self, path: Path, **kwargs):
        super().__init__(is_leaf=path.is_file(), **kwargs)
        self.path = path

    @property
    def label(self) -> str:
        if self.path.is_file():
            prefix = _FILE_PREFIX
        elif self.is_open:
            prefix = _OPEN_FOLDER_PREFIX
        else:
            prefix = _FOLDER_PREFIX
        return f"{_NESTED_PREFIX * self.level}{prefix}{self.path.name}"

    def _toggle_update(self):
        if not self.child_nodes:
            try:
                paths = sorted(
                    self.path.iterdir(), key=lambda path: (path.is_file(), path.name)
                )

                for path in paths:
                    self.add_node(type(self)(path=path))
            except PermissionError:
                return

    def on_mouse(self, mouse_event):
        parent = cast(_FileView, self.parent)
        if (
            mouse_event.nclicks == 2
            and self.is_leaf
            and parent.select_callback is not None
            and parent.selected_node is self
            and self.collides_point(mouse_event.pos)
        ):
            parent.select_callback(self.path)
            return True

        return super().on_mouse(mouse_event)


class _FileView(TreeView[_FileViewNode]):
    def __init__(
        self,
        root_node: _FileViewNode,
        directories_only: bool = False,
        show_hidden: bool = True,
        show_parent_dir: bool = False,
        select_callback: Callable[[Path], None] | None = None,
        filter: Callable[[Path], bool] | None = None,
        **kwargs,
    ):
        self.directories_only = directories_only
        self.show_hidden = show_hidden
        self.show_parent_dir = show_parent_dir
        self.select_callback = select_callback
        self.filter = filter
        super().__init__(root_node=root_node, **kwargs)

    def on_add(self):
        self.update_tree_layout()

    def update_tree_layout(self):
        if self.root is None:
            return

        self.prolicide()

        alpha = self.root_node.alpha
        is_transparent = self.root_node.is_transparent
        it = self.root_node.iter_open_nodes()
        if self.directories_only:
            it = (node for node in it if node.path.is_dir())
        if not self.show_hidden:
            it = (node for node in it if not _is_hidden(node.path))
        if self.filter is not None:
            it = (node for node in it if self.filter(node.path))

        sv = cast(ScrollView, self.parent)
        sv.size = cast(FileChooser, sv.parent).size
        max_width = sv.port_width

        if self.show_parent_dir:
            parent_dir = _ParentDirectory()
            parent_dir.parent_node = self.root_node  # type: ignore
            parent_dir.alpha = alpha
            parent_dir.is_transparent = is_transparent
            max_width = max(max_width, wcswidth(parent_dir.label))
            self.add_gadget(parent_dir)
            start = 1
        else:
            start = 0

        y = 0
        for y, node in enumerate(it, start=start):
            node.alpha = alpha
            node.is_transparent = is_transparent
            node.y = y
            max_width = max(max_width, wcswidth(node.label))
            self.add_gadget(node)
        self.size = y + 1, max_width

        for node in self.children:
            node.size = 1, max_width
            node.add_str(node.label)

    def on_key(self, key_event):
        if not self.children:
            return False

        if key_event.key == "up":
            if self.selected_node is None:
                self.children[0].select()
            else:
                try:
                    index = self.children.index(self.selected_node)
                    if index == 0:
                        index += 1
                except ValueError:
                    index = 1

                self.children[index - 1].select()
        elif key_event.key == "down":
            if self.selected_node is None:
                self.children[0].select()
            else:
                try:
                    index = self.children.index(self.selected_node)
                    if index == len(self.children) - 1:
                        index -= 1
                except ValueError:
                    index = -1
                self.children[index + 1].select()
        elif key_event.key == "left":
            if self.selected_node is None:
                self.children[0].select()
            elif self.selected_node.is_open:
                self.selected_node.toggle()
            elif self.selected_node.parent_node is not self.root_node:
                cast(_FileViewNode, self.selected_node.parent_node).select()
        elif key_event.key == "right":
            if self.selected_node is None:
                self.children[0].select()
            elif self.selected_node.is_leaf:
                pass
            elif not self.selected_node.is_open:
                self.selected_node.toggle()
            elif self.selected_node.child_nodes:
                self.selected_node.child_nodes[0].select()
        elif key_event.key == "enter":
            if self.selected_node is not None and self.select_callback is not None:
                self.select_callback(self.selected_node.path)
        else:
            return super().on_key(key_event)

        if self.selected_node is not None:
            cast(ScrollView, self.parent).scroll_to_rect(self.selected_node.pos)

        return True


class FileChooser(Gadget):
    r"""
    A file chooser gadget.

    Parameters
    ----------
    root_dir : Path | None, default: None
        The root directory of the file chooser or the cwd if not given.
    directories_only : bool, default: False
        Whether to show only directories in the file view.
    show_hidden : bool, default: True
        Whether to show hidden files.
    show_parent_dir : bool, default: False
        Whether parent directory is shown as "..". Double-clicking or pressing enter
        while selected will change the root directory of the file chooser to the parent
        directory.
    select_callback : Callable[[Path], None] | None, default: None
        Called with selected path on double-click or if enter is pressed.
    filter : Callable[[Path], bool] | None, default: None
        Determines whether a path is displayed.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    root_dir : Path
        The root directory of the file chooser.
    directories_only : bool
        Whether to show only directories in the file view.
    show_hidden : bool
        Whether to show hidden files.
    show_parent_dir : bool, default: False
        Whether parent directory is shown as "..".
    select_callback : Callable[[Path], None]
        Called with selected path on double-click or if `enter` is pressed.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        root_dir: Path | None = None,
        directories_only: bool = False,
        show_hidden: bool = True,
        show_parent_dir: bool = False,
        select_callback: Callable[[Path], None] | None = None,
        filter: Callable[[Path], bool] | None = None,
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._root_dir = root_dir or Path()
        self._root_node = _FileViewNode(
            path=self._root_dir, is_transparent=is_transparent
        )
        self._file_view = _FileView(
            root_node=self._root_node,
            directories_only=directories_only,
            show_hidden=show_hidden,
            show_parent_dir=show_parent_dir,
            select_callback=select_callback,
            filter=filter,
            is_transparent=is_transparent,
        )
        self._scroll_view = ScrollView(
            dynamic_bars=True,
            arrow_keys_enabled=False,
            is_transparent=is_transparent,
            alpha=0,
        )
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.alpha = alpha
        self._scroll_view.view = self._file_view
        self.add_gadget(self._scroll_view)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._root_node.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._root_node.alpha = alpha
        for node in self._root_node.iter_open_nodes():
            node.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._root_node.is_transparent = self.is_transparent
        for node in self._root_node.iter_open_nodes():
            node.is_transparent = self.is_transparent
        self._file_view.is_transparent = self.is_transparent
        self._scroll_view.is_transparent = self.is_transparent

    def on_size(self):
        """Update tree layout on resize."""
        self._file_view.update_tree_layout()

    @property
    def directories_only(self):
        """Whether to show only directories in the file view."""
        return self._file_view.directories_only

    @directories_only.setter
    def directories_only(self, directories_only):
        self._file_view.directories_only = directories_only
        self._file_view.update_tree_layout()

    @property
    def show_hidden(self):
        """Whether to show hidden files."""
        return self._file_view.show_hidden

    @show_hidden.setter
    def show_hidden(self, show_hidden):
        self._file_view.show_hidden = show_hidden
        self._file_view.update_tree_layout()

    @property
    def show_parent_dir(self):
        """Whether parent directory is shown as ".."."""
        return self._file_view.show_parent_dir

    @show_parent_dir.setter
    def show_parent_dir(self, show_parent_dir):
        self._file_view.show_parent_dir = show_parent_dir
        self._file_view.update_tree_layout()

    @property
    def root_dir(self) -> Path:
        """The root directory of the file chooser."""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path: Path):
        self._root_dir = path
        if selected := self._file_view.selected_node:
            selected.unselect()
        root = self._root_node
        for node in root.child_nodes:
            node.level = -1
            node.parent_node = None
        root.child_nodes.clear()
        root.is_open = False
        root.path = path
        root.toggle()
        self.vertical_proportion = 0
        self.horizontal_proportion = 0



"""Type annotations for numpy arrays."""

from typing import Literal

from numpy import dtype, float64, intc, ndarray, str_, uint8, uint32, ulong

__all__ = [
    "Cell",
    "Cell0D",
    "Cell1D",
    "Cell2D",
    "Coords",
    "Enum2D",
    "Float1D",
    "Float2D",
    "Int1D",
    "Int2D",
    "RGB_1D",
    "RGB_2D",
    "RGBA_1D",
    "RGBA_2D",
    "RGBM_2D",
    "UInt32_2D",
    "ULong1D",
    "ULong2D",
    "Unicode1D",
    "Unicode2D",
    "braille_pixel_dtype",
    "cell_dtype",
]

cell_dtype = dtype(
    [
        ("ord", "uint32"),
        ("style", "u1"),
        ("fg_color", "u1", (3,)),
        ("bg_color", "u1", (3,)),
    ]
)
"""A structured array type that represents a single cell in a terminal."""

braille_pixel_dtype = dtype(
    [
        ("ord", "uint32"),
        ("total_fg", "double", (4,)),
        ("ncolors", "uint32"),
    ]
)
"""A structured array type for building braille glyphs in a graphic field."""

Cell = ndarray[tuple[int, ...], cell_dtype]
"""An array of ``cell_dtype``."""

Cell0D = ndarray[tuple[()], cell_dtype]
"""A 0-dimensional array of ``cell_dtype``."""

Cell1D = ndarray[tuple[int], cell_dtype]
"""A 1-dimensional array of ``cell_dtype``."""

Cell2D = ndarray[tuple[int, int], cell_dtype]
"""A 2-dimensional array of ``cell_dtype``."""

Float1D = ndarray[tuple[int], dtype[float64]]
"""A 1-dimensional array of floats."""

Float2D = ndarray[tuple[int, int], dtype[float64]]
"""A 2-dimensional array of floats."""

_Coord = ndarray[tuple[Literal[2]] | tuple[Literal[1], Literal[2]], dtype[float64]]
"""2-dimensional coordinates."""

Coords = ndarray[tuple[int, Literal[2]], dtype[float64]]
"""An array of 2-dimensional coordinates."""

Int1D = ndarray[tuple[int], dtype[intc]]
"""A 1-dimensional array of integers."""

Int2D = ndarray[tuple[int, int], dtype[intc]]
"""A 2-dimensional array of integers."""

UInt32_2D = ndarray[tuple[int, int], dtype[uint32]]
"""A 2-dimensional array of unsigned 32-bit integers."""

ULong1D = ndarray[tuple[int], dtype[ulong]]
"""A 1-dimensional array of unsigned long."""

ULong2D = ndarray[tuple[int, int], dtype[ulong]]
"""A 2-dimensional array of unsigned long."""

RGB_1D = ndarray[tuple[int, Literal[3]], dtype[uint8]]
"""A 1-dimensional array of RGB 24-bit colors."""

RGB_2D = ndarray[tuple[int, int, Literal[3]], dtype[uint8]]
"""A 2-dimensional array of RGB 24-bit colors."""

RGBA_1D = ndarray[tuple[int, Literal[4]], dtype[uint8]]
"""A 1-dimensional array of RGBA 32-bit colors."""

RGBA_2D = ndarray[tuple[int, int, Literal[4]], dtype[uint8]]
"""A 2-dimensional array of RGBA 32-bit colors."""

Unicode1D = ndarray[tuple[int], dtype[str_]]
"""A 1-dimensional array of unicode characters."""

Unicode2D = ndarray[tuple[int, int], dtype[str_]]
"""A 2-dimensional array of unicode characters."""

# Rendering array types:

Enum2D = ndarray[tuple[int, int], dtype[uint8]]
"""A 2-dimensional array of bytes used for enumeration."""

RGBM_2D = ndarray[tuple[int, int, Literal[4]], dtype[uint8]]
"""
A 2-dimensional array of RGB 24-bit colors plus a fourth channel ``M`` where non-zero
values indicate opaque pixels and zeros indicate fully transparent pixels.
"""



"""Toggle button behavior for a gadget."""

from collections.abc import Hashable
from typing import Literal, Self
from weakref import WeakValueDictionary

from .button_behavior import ButtonBehavior, ButtonState

__all__ = ["ButtonState", "ToggleButtonBehavior", "ToggleState"]

ToggleState = Literal["on", "off"]
"""Toggle button behavior states."""


class ToggleButtonBehavior(ButtonBehavior):
    """
    Toggle button behavior for gadgets.

    Without a group, a toggle button behaves like a checkbox (either on or off). With a
    group, it behaves like a radio button (only a single button in a group is allowed to
    be in the "on" state).

    Parameters
    ----------
    group : Hashable | None, default: None
        If a group is provided, only one button in a group can be in the on state.
    allow_no_selection : bool, default: False
        If a group is provided, setting this to true allows no selection, i.e.,
        every button can be in the off state.
    always_release : bool, default: False
        Whether a mouse up event outside the button will trigger it.

    Attributes
    ----------
    group : Hashable | None
        If a group is provided, only one button in a group can be in the on state.
    allow_no_selection : bool
        If true and button is in a group, every button can be in the off state.
    toggle_state : ToggleState
        Toggle state of button.
    always_release : bool
        Whether a mouse up event outside the button will trigger it.
    button_state : ButtonState
        Current button state.

    Methods
    -------
    on_toggle()
        Triggled on toggle state change.
    update_off()
        Paint the off state.
    update_on()
        Paint the on state.
    on_release()
        Triggered when a button is released.
    update_normal()
        Paint the normal state.
    update_hover()
        Paint the hover state.
    update_down()
        Paint the down state.
    update_disallowed()
        Paint the disallowed state.
    """

    _toggle_groups: WeakValueDictionary[Hashable, Self] = WeakValueDictionary()

    def __init__(
        self,
        group: Hashable | None = None,
        allow_no_selection: bool = False,
        always_release: bool = False,
        **kwargs,
    ):
        self._toggle_state: ToggleState = "off"
        super().__init__(always_release=always_release, **kwargs)
        self.group = group
        self.allow_no_selection = allow_no_selection

        if (
            group is not None
            and not allow_no_selection
            and ToggleButtonBehavior._toggle_groups.get(group) is None
        ):
            ToggleButtonBehavior._toggle_groups[group] = self
            self._toggle_state = "on"
            self.update_on()
        else:
            self.update_off()

    @property
    def toggle_state(self) -> ToggleState:
        """Toggle state of button."""
        return self._toggle_state

    @toggle_state.setter
    def toggle_state(self, toggle_state: ToggleState):
        if self._toggle_state == toggle_state:
            return

        groups = ToggleButtonBehavior._toggle_groups
        grouped_on = groups.get(self.group)

        self._toggle_state = toggle_state
        if toggle_state == "on":
            if grouped_on is not None:
                grouped_on._toggle_state = "off"
                grouped_on.update_off()
                grouped_on.on_toggle()
            if self.group is not None:
                groups[self.group] = self
            self.update_on()
        else:
            if grouped_on is self:
                if self.allow_no_selection:
                    del groups[self.group]
                else:
                    return
            self.update_off()

        self.on_toggle()

    def on_toggle(self):
        """Update gadget on toggle state change."""

    def on_release(self):
        """Triggered when button is released."""
        self.toggle_state = "off" if self.toggle_state == "on" else "on"

    def update_off(self):
        """Paint the off state."""

    def update_on(self):
        """Paint the on state."""



"""A tabbed gadget."""

from __future__ import annotations

import asyncio
from typing import cast

from ..colors import lerp_colors
from ..text_tools import Style
from .behaviors.themable import Themable
from .behaviors.toggle_button_behavior import ToggleButtonBehavior
from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    new_cell,
)
from .pane import Pane
from .text import Text

__all__ = ["Point", "Size", "Tabs"]

# TODO: Movable tabs?
TAB_SPACING = 3


class _Tab(Themable, ToggleButtonBehavior, Text):
    def __init__(self, title, content, **kwargs):
        super().__init__(**kwargs)
        self.title = title
        self.content = content
        self.set_text(title)

    def on_toggle(self):
        if self.parent is None:
            return
        if self.toggle_state == "on":
            self.content.is_enabled = True

            tabs: Tabs = cast(Tabs, self.parent.parent)

            tabs._history.remove(self.title)
            tabs._history.append(self.title)

            tabs._active_tab = self.title

            if tabs._underline_task is not None:
                tabs._underline_task.cancel()
            tabs._underline_task = asyncio.create_task(
                tabs._tab_underline.tween(
                    duration=0.5,
                    easing="out_bounce",
                    x=self.x - 1,
                    width=self.width + 2,
                )
            )
        else:
            self.content.is_enabled = False

    def _update(self):
        if self.toggle_state == "on":
            fg = self.get_color("titlebar_normal_fg")
            bg = self.get_color("titlebar_normal_bg")
            bold = True
        elif self.button_state == "hover":
            fg = self.hover_color_fg
            bg = self.hover_color_bg
            bold = False
        else:
            fg = self.get_color("titlebar_inactive_fg")
            bg = self.get_color("titlebar_inactive_bg")
            bold = False
        self.canvas["fg_color"] = fg
        self.canvas["bg_color"] = bg
        self.canvas["style"] = Style.BOLD if bold else 0

    update_hover = _update
    update_normal = _update
    update_on = _update
    update_off = _update

    def update_theme(self):
        normal_fg = self.get_color("titlebar_normal_fg")
        normal_bg = self.get_color("titlebar_normal_bg")
        inactive_fg = self.get_color("titlebar_inactive_fg")
        inactive_bg = self.get_color("titlebar_inactive_bg")
        self.hover_color_fg = lerp_colors(normal_fg, inactive_fg, 0.5)
        self.hover_color_bg = lerp_colors(normal_bg, inactive_bg, 0.5)
        self._update()


class Tabs(Themable, Gadget):
    r"""
    A tabbed gadget.

    Parameters
    ----------
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    add_tab(title, content)
        Add a new tab.
    remove_tab(title)
        Remove a tab.
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.tabs: dict[str, _Tab] = {}

        h, w = self.size
        self.tab_bar = Pane(size_hint={"width_hint": 1.0}, size=(1, w))
        self.separator = Text(size_hint={"width_hint": 1.0}, size=(1, w), pos=(1, 0))

        def _update_sep():
            self.separator.chars[:] = "━"

        self.bind("size", _update_sep)

        self.tab_window = Pane(
            size=(h - 2, w),
            pos=(2, 0),
            size_hint={"height_hint": 1.0, "width_hint": 1.0, "height_offset": -2},
        )

        title_fg = self.get_color("titlebar_normal_fg")
        title_bg = self.get_color("titlebar_normal_bg")
        tab_style = {"style": Style.BOLD, "fg_color": title_fg, "bg_color": title_bg}
        self._tab_underline = Text(
            size=(1, 1),
            pos=(1, 0),
            is_enabled=False,
            default_cell=new_cell(ord=ord("━"), **tab_style),
        )
        tab_underline_left = Text(
            size=(1, 1),
            default_cell=new_cell(ord=ord("╺"), **tab_style),
        )
        tab_underline_right = Text(
            size=(1, 1),
            pos_hint={"x_hint": 1.0, "anchor": "right"},
            default_cell=new_cell(ord=ord("╸"), **tab_style),
        )
        self._tab_underline.add_gadgets(tab_underline_left, tab_underline_right)

        self._active_tab: str | None = None
        self._history: list[
            str
        ] = []  # Used to select the last viewed tab when a tab is removed.
        self._underline_task: asyncio.Task | None = None

        self.add_gadgets(
            self.tab_bar, self.separator, self._tab_underline, self.tab_window
        )

    def _reposition_tabs(self):
        x = 1
        for child in self.tab_bar.children:
            child.x = x
            x += child.width + TAB_SPACING

    def update_theme(self):
        """Paint the gadget with current theme."""
        title_inactive_fg = self.get_color("titlebar_inactive_fg")
        title_inactive_bg = self.get_color("titlebar_inactive_bg")
        self.tab_bar.bg_color = title_inactive_bg
        self.tab_window.bg_color = title_inactive_bg
        self.separator.default_cell["fg_color"] = title_inactive_fg
        self.separator.default_cell["bg_color"] = title_inactive_bg
        self.separator.canvas["fg_color"] = title_inactive_fg
        self.separator.canvas["bg_color"] = title_inactive_bg

        title_active_fg = self.get_color("titlebar_normal_fg")
        title_active_bg = self.get_color("titlebar_normal_bg")
        self._tab_underline.default_cell["fg_color"] = title_active_fg
        self._tab_underline.default_cell["bg_color"] = title_active_bg
        self._tab_underline.canvas["fg_color"] = title_active_fg
        self._tab_underline.canvas["bg_color"] = title_active_bg

    def add_tab(self, title: str, content: Gadget):
        """
        Add a new tab. Tab titles are unique.

        Parameters
        ----------
        title : str
            Title of tab.
        content : Gadget
            Content of the tab.
        """
        if title in self.tabs:
            raise ValueError("Tab already exists.")

        tab = _Tab(title, content, group=id(self))
        self.tabs[title] = tab

        self.tab_bar.add_gadget(tab)
        self._reposition_tabs()

        self.tab_window.add_gadget(content)
        self._history.append(title)
        self.tabs[title].toggle_state = "on"
        self._tab_underline.is_enabled = True

    def remove_tab(self, title: str):
        """
        Remove a tab.

        Parameters
        ----------
        title : str
            Title of the tab to remove.
        """
        if title not in self.tabs:
            raise ValueError("Tab doesn't exist.")

        tab = self.tabs.pop(title)
        self.tab_window.remove_gadget(tab.content)
        self.tab_bar.remove_gadget(tab)
        self._history.remove(title)
        self._reposition_tabs()

        if self._active_tab == title:
            if self._history:
                self._active_tab = self._history[-1]
                self.tabs[self._active_tab].toggle_state = "on"
            else:
                self._active_tab = None
                self._tab_underline.is_enabled = False
        elif self._active_tab is not None:
            self._tab_underline.x = self.tabs[self._active_tab].x - 1



cdef struct Band:
    int y1, y2
    size_t size, len
    int *walls


cdef struct CRegion:
    size_t size, len
    Band *bands


cdef class Region:
    cdef CRegion cregion


cdef bint contains(CRegion *cregion, int y, int x)
cdef void bounding_rect(CRegion *cregion, int *y, int *x, size_t *h, size_t *w)



"""
Movable children behavior for a gadget.

Translate movable's children by dragging them.
"""

from typing import Literal

from ...geometry import clamp
from ...terminal.events import MouseButton
from .grabbable import Grabbable

__all__ = ["MovableChildren"]


class MovableChildren(Grabbable):
    """
    Movable children behavior for a gadget.

    Translate a gadget's child by clicking and dragging it.

    Parameters
    ----------
    allow_child_oob : bool, default: True
        Whether child gadgets can be dragged out of parent's bounding box.
    ptf_child_on_grab : bool, default: False
        Whether child gadgets are pulled-to-front when clicked.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"], default: "left"
        Mouse button used for grabbing.

    Attributes
    ----------
    allow_child_oob : bool
        Whether child gadgets can be dragged out of parent's bounding box.
    ptf_child_on_grab : bool
        Whether child gadgets are pulled-to-front when clicked.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"]
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.

    Methods
    -------
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    """

    def __init__(
        self,
        *,
        allow_child_oob=True,
        ptf_child_on_grab=False,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton | Literal["any"] = "left",
        **kwargs,
    ):
        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            **kwargs,
        )
        self.allow_child_oob: bool = allow_child_oob
        """Whether child gadgets can be dragged out of parent's bounding box."""
        self.ptf_child_on_grab: bool = ptf_child_on_grab
        """Whether child gadgets are pulled-to-front when clicked."""
        self._grabbed_child = None

    def grab(self, mouse_event) -> None:
        """Grab the gadget."""
        for child in reversed(self.children):
            if child.collides_point(mouse_event.pos):
                self._is_grabbed = True
                self._grabbed_child = child

                if self.ptf_child_on_grab:
                    child.pull_to_front()

                break
        else:
            super().grab(mouse_event)

    def ungrab(self, mouse_event) -> None:
        """Ungrab the gadget."""
        self._grabbed_child = None
        super().ungrab(mouse_event)

    def grab_update(self, mouse_event) -> None:
        """Update gadget with incoming mouse events while grabbed."""
        if grabbed_child := self._grabbed_child:
            h, w = self.size
            ch, cw = grabbed_child.size
            ct, cl = grabbed_child.pos

            if self.allow_child_oob:
                grabbed_child.top = ct + mouse_event.dy
                grabbed_child.left = cl + mouse_event.dx
            else:
                grabbed_child.top = clamp(ct + mouse_event.dy, 0, h - ch)
                grabbed_child.left = clamp(cl + mouse_event.dx, 0, w - cw)
        else:
            super().grab_update(mouse_event)



"""Base for VT100 terminals."""

from collections.abc import Callable
from typing import Literal

from ..geometry import Point, Size
from .events import Event

DSR_REQUEST_TIMEOUT: float

class Vt100Terminal:
    """
    Base for VT100 terminals.

    Attributes
    ----------
    stdin: int
        The stdin file descriptor.
    stdout: int
        The stdout file descriptor.
    in_alternate_screen : bool
        Whether the alternate screen buffer is enabled.

    Methods
    -------
    process_stdin()
        Read from stdin and feed data into input parser to generate events.
    raw_mode()
        Set terminal to raw mode.
    restore_console()
        Restore console to its original mode.
    attach(event_handler)
        Start generating events from stdin.
    unattach()
        Stop generating events from stdin.
    feed(input_, reset_before)
        Write bytes to stdin parser and return generated events.
    events()
        Return a list of input events and reset the event buffer.
    get_size()
        Get terminal size.
    write(out)
        Write bytes directly to the out-buffer.
    flush()
        Write out-buffer to output stream and flush.
    set_title(title)
        Set terminal title.
    enter_alternate_screen()
        Enter alternate screen buffer.
    exit_alternate_screen()
        Exit alternate screen buffer.
    enable_mouse_support()
        Enable mouse support in terminal.
    disable_mouse_support()
        Disable mouse support in terminal.
    enable_sgr_pixels()
        Enable SGR-PIXELS mouse mode.
    reset_attributes()
        Reset character attributes.
    enable_bracketed_paste()
        Enable bracketed paste in terminal.
    disable_bracketed_paste()
        Disable bracketed paste in terminal.
    show_cursor()
        Show cursor in terminal.
    hide_cursor()
        Hide cursor in terminal.
    enable_reporting_focus()
        Enable reporting terminal focus.
    disable_reporting_focus()
        Disable reporting terminal focus.
    request_cursor_position_report()
        Report current cursor position.
    request_foreground_color()
        Report terminal foreground color.
    request_background_color()
        Report terminal background color.
    request_device_attributes()
        Report device attributes.
    request_pixel_geometry()
        Report pixel geometry per cell.
    request_terminal_geometry()
        Report pixel geometry of terminal.
    request_sgr_pixels_supported()
        Report whether SGR-PIXELS mouse mode is supported.
    request_synchronized_update_mode_supported()
        Report whether synchronized update mode is supported.
    expect_dsr()
        Return whether a device status report is expected.
    move_cursor(pos)
        Move cursor to ``pos``.
    erase_in_display(n)
        Clear part of the screen.
    """

    stdin: int
    stdout: int
    in_alternate_screen: bool
    _event_handler: Callable[[list[Event]], None] | None
    _event_buffer: list[Event]

    def __init__(self, stdin: int = 0, stdout: int = 1) -> None:
        pass

    def process_stdin(self) -> None:
        """Read from stdin and feed data into input parser to generate events."""

    def raw_mode(self) -> None:
        """Set terminal to raw mode."""

    def restore_console(self) -> None:
        """Restore console to its original mode."""

    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        """
        Start generating events from stdin.

        Parameters
        ----------
        event_handler : Callable[[list[Event]], None]
            Callable that handles input events.
        """

    def unattach(self) -> None:
        """Stop generating events from stdin."""

    def feed(self, input_: bytes, reset_before: bool = True) -> list[Event]:
        """
        Write bytes to stdin parser and return generated events.

        Parameters
        ----------
        input_ : bytes
            The bytes to feed to the VT100 parser.
        reset_before : bool, default: True
            Whether to reset the parser state, clear input-buffer, and clear events
            before feeding ``input_``.

        Returns
        -------
        list[Event]
            Events generated from ``input_``.
        """

    def events(self) -> list[Event]:
        """Return a list of input events and reset the event buffer."""

    def get_size(self) -> Size:
        """Get terminal size."""

    def write(self, out: bytes) -> None:
        """Write bytes directly to the out-buffer."""

    def flush(self) -> None:
        """Write out-buffer to output stream and flush."""

    def set_title(self, title: str) -> None:
        """Set terminal title."""

    def enter_alternate_screen(self) -> None:
        """Enter alternate screen buffer."""

    def exit_alternate_screen(self) -> None:
        """Exit alternate screen buffer."""

    def enable_mouse_support(self) -> None:
        """Enable mouse support in terminal."""

    def disable_mouse_support(self) -> None:
        """Disable mouse support in terminal."""

    def enable_sgr_pixels(self) -> None:
        """Enable SGR-PIXELS mouse mode."""

    def disable_sgr_pixels(self) -> None:
        """Disable SGR-PIXELS mouse mode."""

    def reset_attributes(self) -> None:
        """Reset character attributes."""

    def enable_bracketed_paste(self) -> None:
        """Enable bracketed paste in terminal."""

    def disable_bracketed_paste(self) -> None:
        """Disable bracketed paste in terminal."""

    def show_cursor(self) -> None:
        """Show cursor in terminal."""

    def hide_cursor(self) -> None:
        """Hide cursor in terminal."""

    def enable_reporting_focus(self) -> None:
        """Enable reporting terminal focus."""

    def disable_reporting_focus(self) -> None:
        """Disable reporting terminal focus."""

    def request_cursor_position_report(self) -> None:
        """Report current cursor position."""

    def request_foreground_color(self) -> None:
        """Report terminal foreground color."""

    def request_background_color(self) -> None:
        """Report terminal background color."""

    def request_device_attributes(self) -> None:
        """Report device attributes."""

    def request_pixel_geometry(self) -> None:
        """Report pixel geometry per cell."""

    def request_terminal_geometry(self) -> None:
        """Report pixel geometry of terminal."""

    def request_sgr_pixels_supported(self) -> None:
        """Report whether SGR-PIXELS mode is supported."""

    def request_synchronized_update_mode_supported(self) -> None:
        """Report whether synchronized update mode is supported."""

    def expect_dsr(self) -> bool:
        """Return whether a device status report is expected."""

    def move_cursor(self, pos: Point) -> None:
        """
        Move cursor to ``pos``.

        Parameters
        ----------
        pos : Point | None, default: None
            Cursor's new position.
        """

    def erase_in_display(self, n: Literal[0, 1, 2, 3] = 0) -> None:
        """
        Clear part of the screen.

        Parameters
        ----------
        n : int, default: 0
            Determines which part of the screen to clear. If n is ``0``, clear from
            cursor to end of the screen. If n is ``1``, clear from cursor to beginning
            of the screen. If n is ``2``, clear entire screen. If n is ``3``, clear
            entire screen and delete all lines in scrollback buffer.
        """



"""A menu and menu bar gadget."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator
from inspect import signature
from typing import Self, cast

from uwcwidth import wcswidth

from ..geometry import clamp
from ..terminal.events import KeyEvent
from .behaviors.themable import Themable
from .behaviors.toggle_button_behavior import ToggleButtonBehavior, ToggleState
from .grid_layout import GridLayout, Orientation
from .pane import Pane, Point, Pointlike, PosHint, Size, SizeHint, Sizelike
from .text import Text

__all__ = ["Menu", "MenuBar", "MenuDict", "Point", "Size"]

type ItemCallback = Callable[[], None] | Callable[[ToggleState], None]
type MenuDict = dict[tuple[str, str], ItemCallback | MenuDict]
NESTED_SUFFIX = " ▶"
CHECK_OFF = "□"
CHECK_ON = "▣"


def nargs(callable: Callable) -> int:
    """Return the number of arguments of `callable`."""
    return len(signature(callable).parameters)


class _MenuItem(Themable, ToggleButtonBehavior, Pane):
    parent: Menu  # type: ignore

    def __init__(
        self,
        *,
        left_label: str = "",
        right_label: str = "",
        item_callback: ItemCallback = lambda: None,
        submenu: Menu | None = None,
        **kwargs,
    ):
        self.left_label = Text(size=(1, wcswidth(left_label)), alpha=0.0)
        self.right_label = Text(
            size=(1, wcswidth(right_label)),
            pos_hint={"x_hint": 1.0, "anchor": "right"},
            alpha=0.0,
        )
        self.item_callback = item_callback
        self.submenu = submenu
        super().__init__(**kwargs)

        self.left_label.add_str(left_label)
        self.right_label.add_str(right_label)
        self.add_gadgets(self.left_label, self.right_label)
        self.on_transparency()
        self.update_off()

    def _repaint(self):
        if self.button_state == "disallowed":
            fg = self.get_color("menu_item_disallowed_fg")
            bg = self.get_color("menu_item_disallowed_bg")
        elif self.button_state == "normal":
            fg = self.get_color("primary_fg")
            bg = self.get_color("primary_bg")
        else:
            fg = self.get_color("menu_item_hover_fg")
            bg = self.get_color("menu_item_hover_bg")
        self.bg_color = bg
        self.left_label.canvas["fg_color"] = fg
        self.left_label.canvas["bg_color"] = bg
        self.right_label.canvas["fg_color"] = fg
        self.right_label.canvas["bg_color"] = bg

    @property
    def alpha(self) -> float:
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._alpha = alpha
        if self.submenu is not None:
            self.submenu.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self.left_label.is_transparent = self.is_transparent
        self.right_label.is_transparent = self.is_transparent
        if self.submenu is not None:
            self.submenu.is_transparent = self.is_transparent

    def update_theme(self):
        """Paint the gadget with current theme."""
        self._repaint()

    def update_hover(self):
        """Update parent menu and submenu on hover state."""
        self._repaint()

        index = self.parent.children.index(self)
        selected = self.parent._current_selection
        if not (selected == -1 or selected == index):
            self.parent.close_submenus()
            if selected != -1:
                last_hovered = self.parent.children[selected]
                last_hovered.button_state = "normal"

        self.parent._current_selection = index
        if self.submenu is not None:
            self.submenu.open_menu()

    def update_normal(self):
        """Update parent menu and submenu on normal state."""
        self._repaint()
        if self.parent is None:
            return

        if self.submenu is None or not self.submenu.is_enabled:
            if self.parent._current_selection == self.parent.children.index(self):
                self.parent._current_selection = -1
        elif not self.submenu.collides_point(self._last_mouse_pos):
            self.submenu.close_menu()

    def update_disallowed(self):
        self._repaint()
        if self.submenu is not None:
            self.submenu.close_menu()

    def update_off(self):
        """Paint the off state."""
        if self.item_callback is not None and nargs(self.item_callback) == 1:
            self.left_label.chars[0, 1] = CHECK_OFF

    def update_on(self):
        """Paint the on state."""
        if self.item_callback is not None and nargs(self.item_callback) == 1:
            self.left_label.chars[0, 1] = CHECK_ON

    def on_mouse(self, mouse_event):
        """Save last mouse position."""
        self._last_mouse_pos = mouse_event.pos
        return super().on_mouse(mouse_event)

    def on_release(self):
        """Open submenu or call item callback on release."""
        if self.submenu is not None:
            self.submenu.open_menu()
        elif nargs(self.item_callback) == 0:
            self.item_callback()  # type: ignore

            if self.parent.close_on_release:
                self.parent.close_parents()
        else:
            super().on_release()

    def on_toggle(self):
        """Call item callback on toggle state change."""
        if self.item_callback is not None and nargs(self.item_callback) == 1:
            self.item_callback(self.toggle_state)  # type: ignore


class Menu(GridLayout):
    r"""
    A menu gadget.

    Menus are constructed with the class method :meth:`from_dict_of_dicts`. Each key of
    the dict should be a tuple of two strings for left and right labels and each value
    should be either a callable with no arguments for a normal menu item, a callable
    with one argument for a toggle menu item (the argument will be the toggle state of
    the menu item), or a dict (for a submenu).

    Once opened, a menu can be navigated with the mouse or arrow keys.

    Parameters
    ----------
    close_on_release : bool, default: True
        Whether to close the menu when an item is selected.
    close_on_click : bool, default: True
        Whether to close the menu when a click doesn't collide with it.
    alpha : float, default: 1.0
        Transparency of gadget.
    grid_rows : int, default: 1
        Number of rows.
    grid_columns : int, default: 1
        Number of columns.
    orientation : Orientation, default: "tb-lr"
        The orientation of the grid.
    padding_left : int, default: 0
        Padding on left side of grid.
    padding_right : int, default: 0
        Padding on right side of grid.
    padding_top : int, default: 0
        Padding at the top of grid.
    padding_bottom : int, default: 0
        Padding at the bottom of grid.
    horizontal_spacing : int, default: 0
        Horizontal spacing between children.
    vertical_spacing : int, default: 0
        Vertical spacing between children.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    close_on_release : bool
        Whether to close the menu when an item is selected.
    close_on_click : bool
        Whether to close the menu when a click doesn't collide with it.
    alpha : float
        Transparency of gadget.
    min_grid_size : Size
        Minimum grid size needed to show all children.
    grid_rows : int
        Number of rows.
    grid_columns : int
        Number of columns.
    orientation : Orientation
        The orientation of the grid.
    padding_left : int
        Padding on left side of grid.
    padding_right : int
        Padding on right side of grid.
    padding_top : int
        Padding at the top of grid.
    padding_bottom : int
        Padding at the bottom of grid.
    horizontal_spacing : int
        Horizontal spacing between children.
    vertical_spacing : int
        Vertical spacing between children.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    open_menu()
        Open menu.
    close_menu()
        Close menu.
    from_dict_of_dicts(...)
        Constructor to create a menu from a dict of dicts. This should be
        default way of constructing menus.
    index_at(row, col)
        Return index of gadget in :attr:`children` at position `row, col` in the grid.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        close_on_release: bool = True,
        close_on_click: bool = True,
        alpha: float = 1.0,
        orientation: Orientation = "tb-lr",
        grid_rows: int = 1,
        grid_columns: int = 1,
        padding_left: int = 0,
        padding_right: int = 0,
        padding_top: int = 0,
        padding_bottom: int = 0,
        horizontal_spacing: int = 0,
        vertical_spacing: int = 0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            orientation=orientation,
            grid_rows=grid_rows,
            grid_columns=grid_columns,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            padding_bottom=padding_bottom,
            horizontal_spacing=horizontal_spacing,
            vertical_spacing=vertical_spacing,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.close_on_release = close_on_release
        self.close_on_click = close_on_click
        self._alpha = alpha

        self._parent_menu = None
        self._current_selection = -1
        self._submenus = []
        self._menu_button: _MenuButton | None = None

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._alpha = clamp(alpha, 0.0, 1.0)
        for item in self.children:
            item.alpha = self._alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        for item in self.children:
            item.is_transparent = self.is_transparent

    def open_menu(self):
        """Open the menu."""
        self.is_enabled = True

    def close_menu(self):
        """Close the menu."""
        if self._menu_button is not None and self._menu_button.toggle_state == "on":
            self._menu_button.toggle_state = "off"
        else:
            if self._menu_button is not None:
                self._menu_button.update_off()

            self.is_enabled = False
            self._current_selection = -1
            self.close_submenus()

            for child in self.children:
                if child.button_state != "disallowed":
                    child.button_state = "normal"

    def close_submenus(self):
        """Close all submenus."""
        for menu in self._submenus:
            menu.close_menu()

    def close_parents(self):
        """Close all parent menus."""
        if self._parent_menu is not None:
            self._parent_menu.close_parents()
        else:
            self.close_menu()

    def on_mouse(self, mouse_event):
        """Close menus on non-colliding mouse down."""
        if (
            mouse_event.event_type == "mouse_down"
            and self.close_on_click
            and not (
                self.collides_point(mouse_event.pos)
                or any(
                    submenu.collides_point(mouse_event.pos)
                    for submenu in self._submenus
                )
            )
        ):
            self.close_menu()
            return False

        return super().on_mouse(mouse_event)

    def on_key(self, key_event):
        """Navigate menus with arrow keys and select menu items with enter."""
        for submenu in self._submenus:
            if submenu.is_enabled:
                if submenu.on_key(key_event):
                    return True
                else:
                    break

        if key_event.key == "up":
            i = self._current_selection
            if i == -1:
                i = len(self.children) - 1
            else:
                i = self._current_selection
                self.children[i].button_state = "normal"
                i = (i - 1) % len(self.children)

            for _ in self.children:
                if self.children[i].button_state == "disallowed":
                    i = (i - 1) % len(self.children)
                else:
                    self._current_selection = i
                    self.children[i].button_state = "hover"
                    self.close_submenus()
                    return True

            return False

        if key_event.key == "down":
            i = self._current_selection
            if i == -1:
                i = 0
            else:
                self.children[i].button_state = "normal"
                i = (i + 1) % len(self.children)

            for _ in self.children:
                if self.children[i].button_state == "disallowed":
                    i = (i + 1) % len(self.children)
                else:
                    self._current_selection = i
                    self.children[i].button_state = "hover"
                    self.close_submenus()
                    return True

            return False

        if key_event.key == "left":
            if self._current_selection != -1 and (
                (submenu := self.children[self._current_selection].submenu)
                and submenu.is_enabled
            ):
                submenu.close_menu()
                return True

        if key_event.key == "right":
            if self._current_selection != -1 and (
                (submenu := self.children[self._current_selection].submenu)
                and not submenu.is_enabled
            ):
                submenu.open_menu()
                if submenu.children:
                    submenu.children[0].button_state = "hover"
                    submenu.close_submenus()
                return True

        if key_event.key == "enter":
            if (
                self._current_selection != -1
                and (child := self.children[self._current_selection]).submenu is None
            ):
                child.on_release()
                return True

        return super().on_key(key_event)

    @classmethod
    def from_dict_of_dicts(
        cls,
        menu: MenuDict,
        pos: Pointlike = Point(0, 0),
        close_on_release: bool = True,
        close_on_click: bool = True,
        alpha: float = 1.0,
    ) -> Iterator[Self]:
        """
        Create and yield menus from a dict of dicts. Callables should either have no
        arguments for a normal menu item, or one argument for a toggle menu item.

        Parameters
        ----------
        menu : MenuDict
            The menu as a dict of dicts.
        pos : Pointlike, default: Point(0, 0)
            Position of menu.
        close_on_release : bool, default: True
            Whether to close the menu when an item is selected.
        close_on_click : bool, default: True
            Whether to close the menu when a click doesn't collide with it.
        alpha : float, default: 1.0
            Transparency of gadget.

        Yields
        ------
        Menu
            The menu or one of its submenus.
        """
        height = len(menu)
        width = max(
            (
                wcswidth(right_label)
                + wcswidth(left_label)
                + 7
                + isinstance(callable_or_dict, dict) * 2
            )
            for (right_label, left_label), callable_or_dict in menu.items()
        )
        menu_gadget = cls(
            alpha=alpha,
            grid_rows=height,
            grid_columns=1,
            size=(height, width),
            pos=pos,
            close_on_release=close_on_release,
            close_on_click=close_on_click,
        )

        y, x = pos
        for i, ((left_label, right_label), value) in enumerate(menu.items()):
            if isinstance(value, Callable):
                menu_item = _MenuItem(
                    left_label=f"   {left_label}",
                    right_label=f"{right_label} ",
                    item_callback=value,
                    alpha=alpha,
                    size=(1, width),
                )
            elif isinstance(value, dict):
                submenus = cls.from_dict_of_dicts(
                    value,
                    pos=(y + i, x + cast(int, width)),
                    close_on_release=close_on_release,
                    close_on_click=close_on_click,
                    alpha=alpha,
                )
                for submenu in submenus:
                    menu_gadget._submenus.append(submenu)
                    submenu._parent_menu = menu_gadget
                    submenu.is_enabled = False
                    yield submenu

                menu_item = _MenuItem(
                    left_label=f"   {left_label}",
                    right_label=f"{right_label}{NESTED_SUFFIX} ",
                    submenu=submenu,  # type: ignore
                    alpha=alpha,
                    size=(1, width),
                )
            else:
                raise TypeError(f"expected Callable or dict, got {type(value)}")

            menu_gadget.add_gadget(menu_item)

        yield menu_gadget


class _MenuButton(Themable, ToggleButtonBehavior, Text):
    def __init__(self, label, menu, group):
        super().__init__(
            size=(1, wcswidth(label) + 2), group=group, allow_no_selection=True
        )
        self.add_str(f" {label} ")
        self._menu: Menu = menu
        """Menu that the button opens."""

    def _repaint(self):
        if self.button_state == "disallowed":
            fg = self.get_color("menu_item_disallowed_fg")
            bg = self.get_color("menu_item_disallowed_fg")
        elif self.button_state != "normal" or self.toggle_state == "on":
            fg = self.get_color("menu_item_hover_fg")
            bg = self.get_color("menu_item_hover_bg")
        else:
            fg = self.get_color("primary_fg")
            bg = self.get_color("primary_bg")

        self.default_fg_color = fg
        self.default_bg_color = bg
        self.canvas["fg_color"] = fg
        self.canvas["bg_color"] = bg

    def update_theme(self):
        self._repaint()

    def update_down(self):
        self._repaint()

    def update_normal(self):
        self._repaint()

    def update_hover(self):
        self._repaint()
        if self._toggle_groups.get(self.group):
            self.toggle_state = "on"

    def update_on(self):
        self._repaint()

    def update_off(self):
        self._repaint()

    def on_toggle(self):
        if self.toggle_state == "on":
            self._menu.open_menu()
        else:
            self._menu.close_menu()


class MenuBar(GridLayout):
    r"""
    A menu bar.

    A menu bar is constructed with the class method :meth:`from_iterable` from an
    iterable of `(str, MenuDict)`.

    Parameters
    ----------
    close_on_release : bool, default: True
        Whether to close the menu when an item is selected.
    close_on_click : bool, default: True
        Whether to close the menu when a click doesn't collide with it.
    alpha : float, default: 1.0
        Transparency of gadget.
    grid_rows : int, default: 1
        Number of rows.
    grid_columns : int, default: 1
        Number of columns.
    orientation : Orientation, default: "lr-bt"
        The orientation of the grid.
    left_padding : int, default: 0
        Padding on left side of grid.
    right_padding : int, default: 0
        Padding on right side of grid.
    top_padding : int, default: 0
        Padding at the top of grid.
    bottom_padding : int, default: 0
        Padding at the bottom of grid.
    horizontal_spacing : int, default: 0
        Horizontal spacing between children.
    vertical_spacing : int, default: 0
        Vertical spacing between children.
    fg_color : Color | None, default: WHITE
        Foreground color of gadget.
    bg_color : Color | None, default: BLACK
        Background color of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    close_on_release : bool
        Whether to close the menu when an item is selected.
    close_on_click : bool
        Whether to close the menu when a click doesn't collide with it.
    alpha : float
        Transparency of gadget.
    grid_rows : int
        Number of rows.
    grid_columns : int
        Number of columns.
    orientation : Orientation
        The orientation of the grid.
    left_padding : int
        Padding on left side of grid.
    right_padding : int
        Padding on right side of grid.
    top_padding : int
        Padding at the top of grid.
    bottom_padding : int
        Padding at the bottom of grid.
    horizontal_spacing : int
        Horizontal spacing between children.
    vertical_spacing : int
        Vertical spacing between children.
    fg_color : Color | None
        Foreground color of gadget.
    bg_color : Color | None
        Background color of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    open_menu()
        Open menu.
    close_menu()
        Close menu.
    from_dict_of_dicts(...)
        Constructor to create a menu from a dict of dicts. This should be
        default way of constructing menus.
    index_at(row, col)
        Return index of gadget in :attr:`children` at position `row, col` in the grid.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        close_on_release: bool = True,
        close_on_click: bool = True,
        alpha: float = 1.0,
        orientation: Orientation = "tb-lr",
        grid_rows: int = 1,
        grid_columns: int = 1,
        padding_left: int = 0,
        padding_right: int = 0,
        padding_top: int = 0,
        padding_bottom: int = 0,
        horizontal_spacing: int = 0,
        vertical_spacing: int = 0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._close_on_release = close_on_release
        self._close_on_click = close_on_click
        self._alpha = alpha

        super().__init__(
            orientation=orientation,
            grid_rows=grid_rows,
            grid_columns=grid_columns,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            padding_bottom=padding_bottom,
            horizontal_spacing=horizontal_spacing,
            vertical_spacing=vertical_spacing,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

    @property
    def close_on_release(self) -> bool:
        """Whether to close the menu when an item is selected."""
        return self._close_on_release

    @close_on_release.setter
    def close_on_release(self, close_on_release: bool):
        self._close_on_release = close_on_release
        button: _MenuButton
        for button in self.children:
            button._menu.close_on_release = close_on_release

    @property
    def close_on_click(self) -> bool:
        """Whether to close the menu when a click doesn't collide with it."""
        return self._close_on_click

    @close_on_click.setter
    def close_on_click(self, close_on_click: bool):
        self._close_on_click = close_on_click
        button: _MenuButton
        for button in self.children:
            button._menu.close_on_click = close_on_click

    @property
    def alpha(self) -> float:
        """Transparancy of gadget."""
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._alpha = clamp(alpha, 0.0, 1.0)
        button: _MenuButton
        for button in self.children:
            button.alpha = self._alpha
            button._menu.alpha = self._alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        button: _MenuButton
        for button in self.children:
            button.is_transparent = self.is_transparent
            button._menu.is_transparent = self.is_transparent

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Navigate menu bar with left/right arrow keys."""
        button: _MenuButton
        for i, button in enumerate(self.children):
            if button.toggle_state == "on":
                break
        else:
            return super().on_key(key_event)

        if button._menu.on_key(key_event):
            return True

        if key_event.key == "left":
            if self.children[i].button_state != "disallowed":
                self.children[i].button_state = "normal"
            i -= 1
        elif key_event.key == "right":
            if self.children[i].button_state != "disallowed":
                self.children[i].button_state = "normal"
            i += 1
        else:
            return super().on_key(key_event)

        i %= len(self.children)
        # FIXME: Why doesn't `self.children[i].toggle_state = "on"` work?
        self.children[i].button_state = "hover"
        return True

    @classmethod
    def from_iterable(
        cls,
        iter: Iterable[tuple[str, MenuDict]],
        pos: Point = Point(0, 0),
        close_on_release: bool = True,
        close_on_click: bool = True,
        alpha: float = 1.0,
    ) -> Iterator[MenuBar | Menu]:
        """
        Create and yield a menu bar and menus from an iterable of
        `tuple[str, MenuDict]`.

        Parameters
        ----------
        iter : Iterable[tuple[str, MenuDict]]
            An iterable of `tuple[str, MenuDict]` from which to create the menu bar and
            menus.
        pos : Pointlike, default: Point(0, 0)
            Position of menu.
        close_on_release : bool, default: True
            Whether to close the menu when an item is selected.
        close_on_click : bool, default: True
            Whether to close the menu when a click doesn't collide with it.
        alpha : float, default: 1.0
            Transparency of gadget.

        Yields
        ------
        MenuBar | Menu
            A menu or submenu of the menu bar or the menu bar.
        """
        menus = list(iter)

        menubar = cls(
            close_on_release=close_on_release,
            close_on_click=close_on_click,
            alpha=alpha,
            grid_rows=1,
            grid_columns=len(menus),
            size=(1, sum(wcswidth(menu_name) + 2 for menu_name, _ in menus)),
            pos=pos,
        )

        y, x = pos
        for menu_name, menu_dict in menus:
            for menu in Menu.from_dict_of_dicts(
                menu_dict,
                pos=(y + 1, x),
                close_on_release=close_on_release,
                close_on_click=close_on_click,
                alpha=alpha,
            ):
                menu.is_enabled = False
                yield menu

            menu = cast(Menu, menu)  # type: ignore
            menu._menu_button = _MenuButton(
                menu_name, menu, id(menubar)
            )  # Last menu yielded
            menubar.add_gadget(menu._menu_button)
            x += menu._menu_button.width

        yield menubar



"""
A recycle-view provides a view into a large data-set using as few gadgets as
possible.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import Literal

from ..geometry.regions import Region
from ..terminal.events import MouseButton
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike
from .scroll_view import ScrollView

__all__ = ["Point", "RecycleView", "Size"]


class RecycleView[T, G: Gadget](ScrollView):
    r"""
    A recycle-view provides a view into a large data-set using as few gadgets as
    possible.

    ``RecycleView`` is an abstract class. The following methods require an
    implementation:

    * ``get_layout(i)``: Return the size and position of the ith item in
      ``recycle_view_data``.
    * ``new_data_view()``: Initialize and return a new data-view gadget for the recycle
      view.
    * ``update_data_view(data_view, datum)``: Update ``data_view`` to display ``datum``.

    In addition, one may want to re-implement ``set_view_size()`` which sets the size of
    the view of all the data.

    Parameters
    ----------
    recycle_view_data: list[T] | None, default: None
        The recycle-view's data.
    allow_vertical_scroll : bool, default: True
        Allow vertical scrolling.
    allow_horizontal_scroll : bool, default: True
        Allow horizontal scrolling.
    show_vertical_bar : bool, default: True
        Whether the vertical scrollbar is shown.
    show_horizontal_bar : bool, default: True
        Whether the horizontal scrollbar is shown.
    dynamic_bars : bool, default: False
        Whether bars are shown/hidden depending on the view size.
    scrollwheel_enabled : bool, default: True
        Allow vertical scrolling with scrollwheel.
    arrow_keys_enabled : bool, default: True
        Allow scrolling with arrow keys.
    inertial_scrolling_enabled : bool, default: True
        Whether inertial scrolling is enabled.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"], default: "left"
        Mouse button used for grabbing.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    recycle_view_data: list[T]
        The recycle-view's data.
    view : Gadget | None
        The scrolled gadget.
    allow_vertical_scroll : bool
        Allow vertical scrolling.
    allow_horizontal_scroll : bool
        Allow horizontal scrolling.
    show_vertical_bar : bool
        Whether the vertical scrollbar is shown.
    show_horizontal_bar : bool
        Whether the horizontal scrollbar is shown.
    dynamic_bars : bool, default: False
        Whether bars are shown/hidden depending on the view size.
    scrollwheel_enabled : bool
        Allow vertical scrolling with scrollwheel.
    arrow_keys_enabled : bool
        Allow scrolling with arrow keys.
    inertial_scrolling_enabled : bool
        Whether inertial scrolling is enabled.
    vertical_proportion : float
        Vertical scroll position as a proportion of total height.
    horizontal_proportion : float
        Horizontal scroll position as a proportion of total width.
    port_height : int
        Height of view port.
    port_width : int
        Width of view port.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"]
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent : Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    get_layout(i)
        Return the size and position of the ith item of the recycle-view's data.
    new_data_view()
        Initialize and return a new data-view gadget for the recycle-view.
    update_data_view(data_view, datum)
        Update ``data_view`` to display ``datum``.
    set_view_size()
        Set the size of the view.
    refresh_data()
        Refresh data of currently visible data-views.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        recycle_view_data: list[T] | None = None,
        allow_vertical_scroll: bool = True,
        allow_horizontal_scroll: bool = True,
        show_vertical_bar: bool = True,
        show_horizontal_bar: bool = True,
        dynamic_bars: bool = False,
        scrollwheel_enabled: bool = True,
        arrow_keys_enabled: bool = True,
        inertial_scrolling_enabled: bool = True,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton | Literal["any"] = "left",
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            allow_vertical_scroll=allow_vertical_scroll,
            allow_horizontal_scroll=allow_horizontal_scroll,
            show_vertical_bar=show_vertical_bar,
            show_horizontal_bar=show_horizontal_bar,
            dynamic_bars=dynamic_bars,
            scrollwheel_enabled=scrollwheel_enabled,
            arrow_keys_enabled=arrow_keys_enabled,
            inertial_scrolling_enabled=inertial_scrolling_enabled,
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._unused_views: list[G] = []
        """Currently unused views."""
        self._index_to_view: dict[int, G] = {}
        """Index of datum to the view displaying it."""
        self._view_pos_bind: int | None = None
        """uid to unbind view's ``pos`` from ``_update_view``."""
        self._old_len_of_data: int = 0
        """
        Previous length of recycle-view data used to determine whether layout should be
        recalculated in ``refresh_data``.
        """

        self.recycle_view_data: list[T]
        """The recycle-view's data."""
        if recycle_view_data is None:
            self.recycle_view_data = []
        else:
            self.recycle_view_data = recycle_view_data

        self.view: Gadget = Gadget()
        self.set_view_size()
        self.refresh_data()

    @abstractmethod
    def get_layout(self, i: int) -> tuple[Size, Point]:
        """
        Return the size and position of the ith item of the recycle-view's data.

        Parameters
        ----------
        i : int
            The index of an item in the recycle-view's data.

        Returns
        -------
        tuple[Size, Point]
            The size and position of the item.
        """

    @abstractmethod
    def new_data_view(self) -> G:
        """Initialize and return a new data-view gadget for the recycle-view."""

    @abstractmethod
    def update_data_view(self, data_view: G, datum: T) -> None:
        """Update ``data_view`` to display ``datum``."""

    def set_view_size(self) -> None:
        """
        Set the size of the view.

        The default implementation sets the view size to be the minimum size required to
        fully encompass the bottom-most, right-most rect returned by ``get_layout``.
        """
        if self.view is None:
            return

        bottom = right = 0
        for i in range(len(self.recycle_view_data)):
            size, pos = self.get_layout(i)
            current_bottom = size.height + pos.y
            if current_bottom > bottom:
                bottom = current_bottom

            current_right = size.width + pos.x
            if current_right > right:
                right = current_right

        self.view.size = Size(bottom, right)

    def refresh_data(self) -> None:
        """Refresh data of currently visible data-views."""
        if self._old_len_of_data == len(self.recycle_view_data):
            for i, data_view in self._index_to_view.items():
                self.update_data_view(data_view, self.recycle_view_data[i])
        else:
            self._old_len_of_data = len(self.recycle_view_data)
            for data_view in self._index_to_view.values():
                self.view.remove_gadget(data_view)
            self._unused_views.extend(self._index_to_view.values())
            self._index_to_view.clear()
            self.set_view_size()
            self._update_recycle_view()

    def on_add(self) -> None:
        """Bind the view's size and pos to update the items in the view."""
        super().on_add()
        self._view_pos_bind = self.view.bind("pos", self._update_recycle_view)

    def on_remove(self) -> None:
        """Unbind the view's size and pos from updating the items in the view."""
        if self.view is not None and self._view_pos_bind is not None:
            self.view.unbind(self._view_pos_bind)
        self._view_pos_bind = None
        super().on_remove()

    def on_size(self) -> None:
        """Update the recycle-view on resize."""
        super().on_size()
        if self.view is not None:
            self._update_recycle_view()

    def _update_recycle_view(self) -> None:
        scroll_view_reg = Region.from_rect(
            -self.view.pos, (self.port_height, self.port_width)
        )
        recycle_view_reg = Region.from_rect((0, 0), self.view.size)
        clipping_rect = scroll_view_reg & recycle_view_reg

        seen: list[int] = []
        for i in range(len(self.recycle_view_data)):
            size, pos = self.get_layout(i)
            datum_reg = Region.from_rect(pos, size)
            is_seen = bool(datum_reg & clipping_rect)
            has_gadget = i in self._index_to_view
            if has_gadget != is_seen:
                if has_gadget:
                    gadget = self._index_to_view[i]
                    self.view.remove_gadget(gadget)
                    self._unused_views.append(gadget)
                    del self._index_to_view[i]
                else:
                    seen.append(i)

        if len(seen) > len(self._unused_views):
            new_items = len(seen) - len(self._unused_views)
            self._unused_views.extend(self.new_data_view() for _ in range(new_items))

        for i in seen:
            gadget = self._unused_views.pop()
            self._index_to_view[i] = gadget
            gadget.size, gadget.pos = self.get_layout(i)
            self.update_data_view(gadget, self.recycle_view_data[i])
            self.view.add_gadget(gadget)
